"""
@Author = 'Michael Stanley'

============ Change Log ============
2023-Jan-20 = Change license from GPLv2 to GPLv3.

2020-Oct-07 = Created.

============ License ============
Copyright (C) 2020, 2023, 2025 Michael Stanley

This file is part of wmul_rivendell.

wmul_rivendell is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License as published by the Free 
Software Foundation, either version 3 of the License, or (at your option) any 
later version.

wmul_rivendell is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
wmul_rivendell. If not, see <https://www.gnu.org/licenses/>. 
"""
import datetime
import pathlib
import pytest
import wmul_test_utils

from wmul_rivendell import RivendellAudioImporter


@pytest.fixture(scope="function")
def setup_construction(fs):
    mock_file_path = pathlib.Path(r"mock\path\to\file.mp3")
    mock_size = 3_000
    fs.create_file(mock_file_path, st_size=mock_size)
    mock_source_path = pathlib.Path(r"mock\path")
    mock_timestamp = datetime.datetime(year=2020, month=6, day=30, hour=11, minute=5, second=45)

    file_info = RivendellAudioImporter.FileInformation(
        file_path=mock_file_path,
        timestamp=mock_timestamp,
        source_path=mock_source_path
    )

    expected_relative_path = pathlib.Path(r"to\file.mp3")
    expected_str = f"Path: {mock_file_path}, Size: {mock_size}, Timestamp: {mock_timestamp}, " \
                   f"Relative_Path: {expected_relative_path}."
    expected_hash = hash(expected_str)

    return wmul_test_utils.make_namedtuple(
        "setup_construction",
        mock_file_path=mock_file_path,
        mock_size=mock_size,
        mock_timestamp=mock_timestamp,
        file_info=file_info,
        expected_relative_path=expected_relative_path,
        expected_str=expected_str,
        expected_hash=expected_hash
    )


def test_file_path(setup_construction):
    assert setup_construction.mock_file_path == setup_construction.file_info.file_path


def test_file_size(setup_construction):
    assert setup_construction.mock_size == setup_construction.file_info.file_size


def test_timestamp(setup_construction):
    assert setup_construction.mock_timestamp == setup_construction.file_info.timestamp


def test_relative_path(setup_construction):
    assert setup_construction.expected_relative_path == setup_construction.file_info.relative_path


def test_str(setup_construction):
    assert setup_construction.expected_str == str(setup_construction.file_info)


def test_hash(setup_construction):
    assert setup_construction.expected_hash == hash(setup_construction.file_info)


@pytest.fixture(scope="function")
def setup_comparison(fs):
    mock_file_path_string_1 = r"mock\path\to\file1.mp3"
    mock_file_path_1 = pathlib.Path(mock_file_path_string_1)
    mock_size_1 = 3_000
    fs.create_file(mock_file_path_string_1, st_size=mock_size_1)
    mock_source_path_1 = pathlib.Path(r"mock\path")
    mock_timestamp_1 = datetime.datetime(year=2020, month=6, day=30, hour=11, minute=5, second=45)

    mock_file_path_string_3 = r"mock\path\to\file3.mp3"
    mock_file_path_3 = pathlib.Path(mock_file_path_string_3)
    fs.create_file(mock_file_path_string_3, st_size=mock_size_1)

    file_info_1 = RivendellAudioImporter.FileInformation(
        file_path=mock_file_path_1,
        timestamp=mock_timestamp_1,
        source_path=mock_source_path_1
    )

    file_info_2 = RivendellAudioImporter.FileInformation(
        file_path=mock_file_path_1,
        timestamp=mock_timestamp_1,
        source_path=mock_source_path_1
    )

    file_info_3 = RivendellAudioImporter.FileInformation(
        file_path=mock_file_path_3,
        timestamp=mock_timestamp_1,
        source_path=mock_source_path_1
    )

    mock_size_4 = 4_000
    fs.remove_object(mock_file_path_string_1)
    fs.create_file(mock_file_path_string_1, st_size=mock_size_4)

    file_info_4 = RivendellAudioImporter.FileInformation(
        file_path=mock_file_path_1,
        timestamp=mock_timestamp_1,
        source_path=mock_source_path_1
    )

    mock_timestamp_5 = datetime.datetime(year=2020, month=6, day=30, hour=11, minute=10, second=45)

    file_info_5 = RivendellAudioImporter.FileInformation(
        file_path=mock_file_path_1,
        timestamp=mock_timestamp_5,
        source_path=mock_source_path_1
    )

    return wmul_test_utils.make_namedtuple(
        "setup_comparison",
        file_info_1=file_info_1,
        file_info_2=file_info_2,
        file_info_3=file_info_3,
        file_info_4=file_info_4,
        file_info_5=file_info_5
    )


def test_equality(setup_comparison):
    assert setup_comparison.file_info_1 == setup_comparison.file_info_2
    assert setup_comparison.file_info_1 != setup_comparison.file_info_3
    assert setup_comparison.file_info_1 != setup_comparison.file_info_4
    assert setup_comparison.file_info_1 != setup_comparison.file_info_5

    assert setup_comparison.file_info_2 != setup_comparison.file_info_3
    assert setup_comparison.file_info_2 != setup_comparison.file_info_4
    assert setup_comparison.file_info_2 != setup_comparison.file_info_5

    assert setup_comparison.file_info_3 != setup_comparison.file_info_4
    assert setup_comparison.file_info_3 != setup_comparison.file_info_5

    assert setup_comparison.file_info_4 != setup_comparison.file_info_5


def test_less_than(setup_comparison):
    assert not(setup_comparison.file_info_1 < setup_comparison.file_info_2)
    assert setup_comparison.file_info_1 < setup_comparison.file_info_3
    assert setup_comparison.file_info_1 < setup_comparison.file_info_4
    assert setup_comparison.file_info_1 < setup_comparison.file_info_5

    assert not(setup_comparison.file_info_2 < setup_comparison.file_info_1)
    assert setup_comparison.file_info_2 < setup_comparison.file_info_3
    assert setup_comparison.file_info_2 < setup_comparison.file_info_4
    assert setup_comparison.file_info_2 < setup_comparison.file_info_5

    assert not(setup_comparison.file_info_3 < setup_comparison.file_info_1)
    assert not(setup_comparison.file_info_3 < setup_comparison.file_info_2)
    assert not(setup_comparison.file_info_3 < setup_comparison.file_info_4)
    assert not(setup_comparison.file_info_3 < setup_comparison.file_info_5)

    assert not(setup_comparison.file_info_4 < setup_comparison.file_info_1)
    assert not(setup_comparison.file_info_4 < setup_comparison.file_info_2)
    assert setup_comparison.file_info_4 < setup_comparison.file_info_3
    assert setup_comparison.file_info_4 < setup_comparison.file_info_5


@pytest.fixture(scope="function")
def setup_same_size_as(mocker):
    size_1 = 3_000
    size_2 = 4_000
    mock_info_1_a = mocker.Mock(
        file_size=size_1
    )
    mock_info_1_b = mocker.Mock(
        file_size=size_1
    )
    mock_info_2 = mocker.Mock(
        file_size=size_2
    )

    return wmul_test_utils.make_namedtuple(
        "setup_same_size_as",
        mock_info_1_a=mock_info_1_a,
        mock_info_1_b=mock_info_1_b,
        mock_info_2=mock_info_2
    )


def test_same_size_true(setup_same_size_as):
    assert RivendellAudioImporter.FileInformation.same_size_as(setup_same_size_as.mock_info_1_a,
                                                               setup_same_size_as.mock_info_1_a)
    assert RivendellAudioImporter.FileInformation.same_size_as(setup_same_size_as.mock_info_1_b,
                                                               setup_same_size_as.mock_info_1_b)
    assert RivendellAudioImporter.FileInformation.same_size_as(setup_same_size_as.mock_info_2,
                                                               setup_same_size_as.mock_info_2)
    assert RivendellAudioImporter.FileInformation.same_size_as(setup_same_size_as.mock_info_1_a,
                                                               setup_same_size_as.mock_info_1_b)
    assert RivendellAudioImporter.FileInformation.same_size_as(setup_same_size_as.mock_info_1_b,
                                                               setup_same_size_as.mock_info_1_a)


def test_same_size_false(setup_same_size_as):
    assert not RivendellAudioImporter.FileInformation.same_size_as(setup_same_size_as.mock_info_1_a,
                                                                   setup_same_size_as.mock_info_2)
    assert not RivendellAudioImporter.FileInformation.same_size_as(setup_same_size_as.mock_info_1_b,
                                                                   setup_same_size_as.mock_info_2)
    assert not RivendellAudioImporter.FileInformation.same_size_as(setup_same_size_as.mock_info_2,
                                                                   setup_same_size_as.mock_info_1_a)
    assert not RivendellAudioImporter.FileInformation.same_size_as(setup_same_size_as.mock_info_2,
                                                                   setup_same_size_as.mock_info_1_b)


@pytest.fixture(scope="function")
def setup_new_info_much_newer(mocker):
    five_ten_timestamp = datetime.datetime(year=2020, month=6, day=30, hour=12, minute=5, second=10)
    five_thirty_timestamp = datetime.datetime(year=2020, month=6, day=30, hour=12, minute=5, second=30)
    five_fifty_timestamp = datetime.datetime(year=2020, month=6, day=30, hour=12, minute=5, second=50)

    five_ten_info = mocker.Mock(
        timestamp=five_ten_timestamp
    )
    five_thirty_info = mocker.Mock(
        timestamp=five_thirty_timestamp
    )
    five_fifty_info = mocker.Mock(
        timestamp=five_fifty_timestamp
    )

    minimum_difference = 30

    return wmul_test_utils.make_namedtuple(
        "setup_new_info_much_newer",
        five_ten_info=five_ten_info,
        five_thirty_info=five_thirty_info,
        five_fifty_info=five_fifty_info,
        minimum_difference=minimum_difference
    )


def test_five_ten_and_five_thirty(setup_new_info_much_newer):
    assert not RivendellAudioImporter.FileInformation.new_info_much_newer(
        setup_new_info_much_newer.five_ten_info,
        setup_new_info_much_newer.five_thirty_info,
        setup_new_info_much_newer.minimum_difference
    )

    assert not RivendellAudioImporter.FileInformation.new_info_much_newer(
        setup_new_info_much_newer.five_thirty_info,
        setup_new_info_much_newer.five_ten_info,
        setup_new_info_much_newer.minimum_difference
    )


def test_five_ten_and_five_fifty(setup_new_info_much_newer):
    assert RivendellAudioImporter.FileInformation.new_info_much_newer(
        setup_new_info_much_newer.five_ten_info,
        setup_new_info_much_newer.five_fifty_info,
        setup_new_info_much_newer.minimum_difference
    )

    assert not RivendellAudioImporter.FileInformation.new_info_much_newer(
        setup_new_info_much_newer.five_fifty_info,
        setup_new_info_much_newer.five_ten_info,
        setup_new_info_much_newer.minimum_difference
    )


@pytest.fixture(scope="function", params=[False, True], ids=["No Log Argument", "Has Log Argument"])
def setup_generate_importer_command_no_group(fs, caplog, request):
    has_log_argument = request.param
    mock_file_path = pathlib.Path(r"/source/mock_filename.mp3")
    fs.create_file(mock_file_path)
    mock_time_stamp = datetime.datetime.now()
    mock_source_path = pathlib.Path(r"/source")

    mock_file_for_importer = RivendellAudioImporter.FileInformation(
        mock_file_path, mock_time_stamp, mock_source_path
    )

    if has_log_argument:
        mock_log_argument = "mock_log_argument"
    else:
        mock_log_argument = ""

    result_value = mock_file_for_importer.generate_importer_command(mock_log_argument)

    expected_log_messages = [
        f"With {mock_file_for_importer}",
        f"This file does not have a group. {str(mock_file_path)}"
    ]

    return wmul_test_utils.make_namedtuple(
        "setup_generate_importer_command_no_group",
        result_log_messages=caplog.messages,
        result_value=result_value,
        expected_log_messages=expected_log_messages
    )


def test_generate_importer_command_no_group_log_message_correct(setup_generate_importer_command_no_group):
    assert sorted(setup_generate_importer_command_no_group.expected_log_messages) == \
           sorted(setup_generate_importer_command_no_group.result_log_messages)


def test_importer_command_no_group_return_value_correct(setup_generate_importer_command_no_group):
    assert setup_generate_importer_command_no_group.result_value is None


@pytest.fixture(scope="function", params=[False, True], ids=["No Log Argument", "Has Log Argument"])
def setup_generate_importer_command_group_no_sched_code(fs, caplog, request):
    has_log_argument = request.param
    mock_group = "mock_group"
    mock_file_path_str = f"\\source\\{mock_group}\\mock_filename.mp3"
    mock_file_path = pathlib.Path(mock_file_path_str)
    fs.create_file(mock_file_path)
    mock_time_stamp = datetime.datetime.now()
    mock_source_path = pathlib.Path(r"/source")

    mock_file_for_importer = RivendellAudioImporter.FileInformation(
        mock_file_path, mock_time_stamp, mock_source_path
    )

    if has_log_argument:
        mock_log_argument = "mock_log_argument"
        expected_result = [
            "--autotrim-level=0",
            "--normalization-level=0",
            "--title-from-cartchunk-cutid",
            "--delete-source",
            f'--set-string-description="mock_filename.mp3"',
            '--verbose',
            mock_log_argument,
            mock_group,
            mock_file_path_str
        ]
    else:
        mock_log_argument = ""
        expected_result = [
            "--autotrim-level=0",
            "--normalization-level=0",
            "--title-from-cartchunk-cutid",
            "--delete-source",
            f'--set-string-description="mock_filename.mp3"',
            '--verbose',
            mock_group,
            mock_file_path_str
        ]

    result_value = mock_file_for_importer.generate_importer_command(mock_log_argument)

    expected_log_messages = [
        f"With {mock_file_for_importer}",
        "No scheduler codes."
    ]

    return wmul_test_utils.make_namedtuple(
        "setup_generate_importer_command_group_no_sched_code",
        result_log_messages=caplog.messages,
        expected_log_messages=expected_log_messages,
        result_value=result_value,
        expected_result=expected_result,
    )


def test_file_information_generate_importer_command_group_no_sched_code_log_message_correct(
        setup_generate_importer_command_group_no_sched_code):
    assert sorted(setup_generate_importer_command_group_no_sched_code.expected_log_messages) == \
           sorted(setup_generate_importer_command_group_no_sched_code.result_log_messages)


def test_file_information_generate_importer_command_group_return_value_correct(
        setup_generate_importer_command_group_no_sched_code):

    assert setup_generate_importer_command_group_no_sched_code.result_value == \
           setup_generate_importer_command_group_no_sched_code.expected_result


@pytest.fixture(scope="function", params=[False, True], ids=["No Log Argument", "Has Log Argument"])
def setup_generate_importer_command_group_sched_codes(fs, caplog, request):
    has_log_argument = request.param
    mock_group = "mock_group"
    mock_sched_code_1 = "mock_sched_code_1"
    mock_sched_code_2 = "mock_sched_code_2"
    mock_file_path_str = f"\\source\\{mock_group}\\{mock_sched_code_1}\\{mock_sched_code_2}\\mock_filename.mp3"
    mock_file_path = pathlib.Path(mock_file_path_str)
    fs.create_file(mock_file_path)
    mock_time_stamp = datetime.datetime.now()
    mock_source_path = pathlib.Path(r"/source")

    mock_file_for_importer = RivendellAudioImporter.FileInformation(
        mock_file_path, mock_time_stamp, mock_source_path
    )

    if has_log_argument:
        mock_log_argument = "mock_log_arguemnt"
        expected_result = [
            f"--add-scheduler-code={mock_sched_code_1}",
            f"--add-scheduler-code={mock_sched_code_2}",
            "--autotrim-level=0",
            "--normalization-level=0",
            "--title-from-cartchunk-cutid",
            "--delete-source",
            f'--set-string-description="mock_filename.mp3"',
            "--verbose",
            mock_log_argument,
            mock_group,
            mock_file_path_str
        ]
    else:
        mock_log_argument = ""
        expected_result = [
            f"--add-scheduler-code={mock_sched_code_1}",
            f"--add-scheduler-code={mock_sched_code_2}",
            "--autotrim-level=0",
            "--normalization-level=0",
            "--title-from-cartchunk-cutid",
            "--delete-source",
            f'--set-string-description="mock_filename.mp3"',
            "--verbose",
            mock_group,
            mock_file_path_str
        ]

    result_value = mock_file_for_importer.generate_importer_command(mock_log_argument)

    expected_log_messages = [
        f"With {mock_file_for_importer}",
        "Has scheduler codes."
    ]

    return wmul_test_utils.make_namedtuple(
        "setup_generate_importer_command_group_sched_codes",
        result_log_messages=caplog.messages,
        expected_log_messages=expected_log_messages,
        result_value=result_value,
        expected_result=expected_result
    )


def test_generate_importer_command_group_sched_codes_log_message_correct(
        setup_generate_importer_command_group_sched_codes):
    assert sorted(setup_generate_importer_command_group_sched_codes.expected_log_messages) == \
           sorted(setup_generate_importer_command_group_sched_codes.result_log_messages)


def test_generate_importer_command_group_sched_codes_return_value_correct(
        setup_generate_importer_command_group_sched_codes):

    assert setup_generate_importer_command_group_sched_codes.result_value == \
           setup_generate_importer_command_group_sched_codes.expected_result


def test__failed(fs):
    mock_file_path = pathlib.Path(r"\source\mock_group\mock_filename.mp3")
    fs.create_file(mock_file_path)
    mock_time_stamp = datetime.datetime.now()
    mock_source_path = pathlib.Path(r"\source")

    mock_file_info = RivendellAudioImporter.FileInformation(
        mock_file_path, mock_time_stamp, mock_source_path
    )

    mock_suffix = "_mock_suffix"

    mock_file_info._failed(mock_suffix)

    expected_file = pathlib.Path(r"\source\mock_group\mock_filename.mp3_mock_suffix")
    assert expected_file.exists()
    assert not mock_file_path.exists()


@pytest.fixture(scope="function")
def setup_failed_with_reason(fs, mocker):
    mock_failed = mocker.Mock()

    mocker.patch("wmul_rivendell.RivendellAudioImporter.FileInformation._failed",
                 mock_failed)

    mock_file_path = pathlib.Path(r"\source\mock_group\mock_filename.mp3")
    fs.create_file(mock_file_path)
    mock_time_stamp = datetime.datetime.now()
    mock_source_path = pathlib.Path(r"\source")

    mock_file_info = RivendellAudioImporter.FileInformation(
        mock_file_path, mock_time_stamp, mock_source_path
    )

    return wmul_test_utils.make_namedtuple(
        "setup_failed_with_reason",
        mock_file_info=mock_file_info,
        mock_failed=mock_failed
    )


def test_failed_rdimport(setup_failed_with_reason):
    setup_failed_with_reason.mock_file_info.failed_rdimport()
    setup_failed_with_reason.mock_failed.assert_called_once_with("_RDIMPORT_FAILED")


def test_failed_group(setup_failed_with_reason):
    setup_failed_with_reason.mock_file_info.failed_group()
    setup_failed_with_reason.mock_failed.assert_called_once_with("_NO_GROUP_FAILED")



"""
@Author = 'Michael Stanley'

Command-Line-Interface for the various modules in this package.

============ Change Log ============
2025-Jan-29 = Add option to database_statistics for write_full_statistics option on DatabaseStatistics.

2025-Jan-23 = Add options to database_statistics for smallest_stdev, minimum_population, upper and lower bound 
              multiples. 

2025-Jan-03 = Add database_statistics option.
              Fix type in help text of filter_cart_report.

2023-May-25 = Refactor filter_cart_report to reflect that the loading of the
                cart data from a file has been refactored into LoadCartDataDump.

2023-Jan-20 = Change license from GPLv2 to GPLv3.
              Back-insert items into change log.
              Update description.

2023-Jan-19 = Add --fix_header to filter_cart_report. 
              Change the cli options to use the versions of RequiredIf and 
                MXWith now in wmul_click_utils.

2022-May-16 = Fix cli help for load_current_log_line.

2022-May-06 = Update documentation.

2020-Oct-15 = Add current version number to logging.
              Add ability for RivendellAudioImporter to tell rdimport where to
                log.
              Log final crashes of RivendellAudioImporter.

2020-Jun-26 = Created.

============ License ============
Copyright (C) 2020, 2022-2023, 2025 Michael Stanley

This file is part of wmul_rivendell.

wmul_rivendell is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License as published by the Free 
Software Foundation, either version 3 of the License, or (at your option) any 
later version.

wmul_rivendell is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
wmul_rivendell. If not, see <https://www.gnu.org/licenses/>. 
"""
import click
import datetime

from pathlib import Path
from wmul_rivendell.DatabaseStatistics import DatabaseStatistics, StatisticsLimits
from wmul_rivendell.FilterCartReportForMusicScheduler import FilterCartReportForMusicScheduler
from wmul_rivendell.LoadCartDataDump import LoadCartDataDump
from wmul_rivendell.LoadCurrentLogLine import LoadCurrentLogLineArguments, run_script as load_current_log_lines
from wmul_rivendell.RivendellAudioImporter import \
    ImportRivendellFileWithFileSystemMetadataArguments, run_script as import_rivendell_file
from wmul_click_utils import RequiredIf, MXWith

import wmul_emailer
import wmul_logger


_logger = wmul_logger.get_logger()


@click.group()
@click.version_option()
@click.option('--log_name', type=click.Path(exists=False, file_okay=True, dir_okay=False, writable=True), default=None,
              required=False, help="The path to the log file.")
@click.option('--log_level', type=click.IntRange(10, 50, clamp=True), required=False, default=30,
              help="The log level: 10: Debug, 20: Info, 30: Warning, 40: Error, 50: Critical. "
                   "Intermediate values (E.G. 32) are permitted, but will essentially be rounded up (E.G. Entering 32 "
                   "is the same as entering 40. Logging messages lower than the log level will not be written to the "
                   "log. E.G. If 30 is input, then all Debug, Info, and Verbose messages will be silenced.")
def wmul_rivendell_cli(log_name, log_level):
    if log_name:
        global _logger
        _logger = wmul_logger.setup_logger(file_name=log_name, log_level=log_level)
        import pkg_resources  # part of setuptools
        version = pkg_resources.require("wmul_rivendell")[0].version
        _logger.warning(f"Version: {version}")
        _logger.warning("In command_line_interface")


@wmul_rivendell_cli.command()
@click.argument('rivendell_cart_filename', type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True),
                nargs=1)
@click.argument('output_filename', type=click.Path(exists=False, file_okay=True, dir_okay=False, writable=True),
                nargs=1)
@click.option('--include_all_cuts', is_flag=True,
              help="Whether to use every cut in the cart when making the calculations. If this is set, every cut from "
              "each cart will be used in the calculations. If it is not set, only the lowest numbered cut will be " 
              "used.")
@click.option('--excluded_groups_file_name', type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True),
              help="File path to a text file containing a list of groups to be exluded. The file should have one "
              "group name on each line. A group name may be present in this file but not in the cart data dump.")
@click.option('--smallest_stdev', type=int, default=15, help="The smallest standard deviation (in seconds) permitted for "
              "calculating outliers, and upper and lower bounds for excluding songs. If the group population standard "
              "deviation is less than this number, then no outliers will be excluded. If the population (excluding "
              "outliers) standard deviation is less than this number, then no lower and upper boundes will be "
              "calculated. This limit helps avoid unwanted behaviour when a group is mostly the same length. E.G. a "
              "group containing only 30 second spots.")
@click.option("--minimum_population", type=int, default=4, help="The smallest population for calculating outliers. If "
              "the population of the group is smaller than or equal to this number, then outliers will not be " 
              "calculated and excluded.")
@click.option("--lower_bound_multiple", type=float, default=1.5, help="The standard deviation will be multiplied by " 
              "this number and subtracted from the mean to generate the lower bound.")
@click.option("--upper_bound_multiple", type=float, default=3.0, help="The standard deviation will be multiplied by " 
              "this number and added to the mean to generate the upper bound.")
@click.option("--write_limits", is_flag=True, help="Whether to write the statistics limits (smallest_stdev, "
              "minimum_population, lower_bound_multiple, and upper_bound_multiple) to the file.")
@click.option("--write_full_statistics", is_flag=True, help="If true, the full set of statistics will be written. " 
              "If false, only the summary statistics (Number of Songs, Lower Bound, Upper Bound) will be written.")
def database_statistics(rivendell_cart_filename, output_filename, include_all_cuts, excluded_groups_file_name, 
                        smallest_stdev, minimum_population, lower_bound_multiple, upper_bound_multiple, write_limits,
                        write_full_statistics):
    _logger.debug(f"With {locals()}")

    stats_limits = StatisticsLimits(
        smallest_stdev=smallest_stdev,
        minimum_population_for_outliers=minimum_population,
        lower_bound_multiple=lower_bound_multiple,
        upper_bound_multiple=upper_bound_multiple
    )

    excluded_groups = get_excluded_groups(excluded_groups_file_name)
    output_filename = Path(output_filename)

    lcdd = LoadCartDataDump(
        rivendell_cart_data_filename=rivendell_cart_filename,
        include_all_cuts=include_all_cuts,
        include_macros=False,
        excluded_group_list=excluded_groups
    )

    rivendell_carts = lcdd.load_carts()

    x = DatabaseStatistics(
        rivendell_carts=rivendell_carts,
        output_filename=output_filename,
        stats_limits=stats_limits,
        write_limits=write_limits,
        write_full_statistics=write_full_statistics
    )
    x.run_script()


@wmul_rivendell_cli.command()
@click.argument('rivendell_cart_filename', type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True),
                nargs=1)
@click.argument('output_filename', type=click.Path(exists=False, file_okay=True, dir_okay=False, writable=True),
                nargs=1)
@click.argument('desired_fields_filename', type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True),
                nargs=1)
@click.option('--include_macros', is_flag=True, help="Include macro carts in the output file.")
@click.option('--include_all_cuts', is_flag=True,
              help="Include all the individual cuts in the output file. If this is left unset, then the output file "
                   "will only include the lowest numbered cut in each cart.")
@click.option('--excluded_groups_file_name', 
              type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True),
              help="File path to a text file containing a list of groups to be exluded. The file should have one "
              "group name on each line. A group name may be present in this file but not in the cart data dump.")
@click.option('--use_trailing_comma', is_flag=True,
              help="Include a comma at the end of each line. Required by some scheduling software, such as Natural "
                   "Music, to see the final field.")
def filter_cart_report(rivendell_cart_filename, output_filename, desired_fields_filename, include_macros,
                       include_all_cuts, excluded_groups_file_name, use_trailing_comma):
    _logger.debug(f"With {locals()}")

    with open(desired_fields_filename, "rt") as desired_fields_reader:
        desired_fields = [desired_field.strip("\n\r") for desired_field in desired_fields_reader]

    excluded_groups = get_excluded_groups(excluded_groups_file_name)

    lcdd = LoadCartDataDump(
        rivendell_cart_data_filename=rivendell_cart_filename,
        include_macros=include_macros,
        include_all_cuts=include_all_cuts,
        excluded_group_list=excluded_groups,
    )

    rivendell_carts = lcdd.load_carts()

    x = FilterCartReportForMusicScheduler(
        rivendell_carts=rivendell_carts,
        output_filename=output_filename,
        desired_field_list=desired_fields,
        use_trailing_comma=use_trailing_comma
    )
    x.run_script()


@wmul_rivendell_cli.command()
@click.argument('log_name_format', type=str, nargs=1)
@click.argument('rivendell_host', type=str, nargs=1)
@click.option('--sql_host', type=str, nargs=1, default="localhost",
              help="The host name to the SQL database. Usually localhost. Default: localhost.")
@click.option('--sql_user', type=str, nargs=1, default="rduser",
              help="The username for the SQL database. Usually rduser. Default: rduser.")
@click.option('--sql_pass', type=str, nargs=1, default="letmein",
              help="The password for the SQL database. Usually letmein. Default: letmein.")
@click.option('--sql_database_name', type=str, nargs=1, default="Rivendell",
              help="The Database name of the SQL database. Usually Rivendell. Default: Rivendell.")
@click.option('--use_date', type=click.DateTime(formats=["%Y-%m-%d", "%y-%m-%d"]), nargs=1,
              help="The date of the log to be loaded. Format is YY-MM-DD or YYYY-MM-DD. If this option is omitted, "
                   "the system date of the system running the script will be used.")
@click.option('--use_time',
              type=click.DateTime(formats=["%I:%M:%S %p", "%I:%M %p", "%I %p", "%H:%M:%S", "%H:%M", "%H"]),
              nargs=1, help="The time of the log line to be loaded. The script will find the line closest to, but "
                            "before that time. Valid formats are HH:MM:SS AM, HH:MM AM, HH AM, HH:MM:SS, HH:MM, and HH."
                            "If AM/PM are present, HH will be 12-hour. If AM/PM are absent, HH will be 24-hour. IF MM "
                            "and/or SS are omitted, they will be set to 00. If this option is omitted,"
                            "the system time of the system running the script will be used.")
@click.option('--dry_run', is_flag=True,
              help="For testing purposes. Prints out the log line that is selected, but does not load it.")
@click.option('--start_immediately', is_flag=True,
              help="Starts the selected log line immediately. If not set, the selected log line will be 'made next'.")
@click.option('--days_back', type=int, nargs=1, default=7,
              help="Maximum number of days back in time to go. If a log is not available for the given day, the script"
                   " will try to load the previous day's log. It will keep going back in time up to and including "
                   "this many days. This option is for cases where it is preferred to load and replay an old log "
                   "rather than no log. If no logs can be found for those dates, it will try to load the default log, "
                   "if provided. Set this value to 0 to not attempt previous days' logs. Defaults to 7.")
@click.option('--default_log', type=str, nargs=1,
              help="The full name of the last-ditch log to try to load if day based logs fail.")
@click.option('--log_machine', type=int, nargs=1, default=1,
              help="The log machine on which to load the playlist. Defaults to 1 (Main Log).")
@click.option("--email_address", type=str, multiple=True,
              help="The e-mail address to which the report should be sent.")
@click.option("--mail_server", type=str, cls=RequiredIf, required_if="email_address",
              help="The address of the e-mail SMTP server to use.")
@click.option("--mail_port", type=int, default=25, help="The port of the e-mail server. Defaults to 25")
@click.option("--mail_username", type=str, help="The username to authenticate with the e-mail server.")
@click.option("--mail_password", type=str, help="The password to authenticate with the e-mail server.")
def load_current_log_line(log_name_format, rivendell_host, sql_host, sql_user, sql_pass, sql_database_name, use_date,
                          use_time, dry_run, start_immediately, days_back, default_log, log_machine, email_address,
                          mail_server, mail_port, mail_username, mail_password):
    _logger.debug(f"With {locals()}")

    if email_address:
        emailer = wmul_emailer.EmailSender(
            server_host=mail_server,
            port=mail_port,
            user_name=mail_username,
            password=mail_password,
            destination_email_addresses=email_address,
            from_email_address='rivendell@wmul-nas-4'
        )
    else:
        emailer = None

    if use_date:
        use_date = use_date.date()
    else:
        use_date = datetime.datetime.now().date()

    if use_time:
        use_time = use_time.time()
    else:
        use_time = datetime.datetime.now().time()

    use_datetime = datetime.datetime.combine(date=use_date, time=use_time)

    x = LoadCurrentLogLineArguments(
        log_datetime=use_datetime,
        rivendell_host=rivendell_host,
        sql_host=sql_host,
        sql_user=sql_user,
        sql_pass=sql_pass,
        sql_database_name=sql_database_name,
        log_name_format=log_name_format,
        dry_run=dry_run,
        start_immediately=start_immediately,
        days_back=days_back,
        default_log=default_log,
        log_machine=log_machine,
        emailer=emailer
    )

    load_current_log_lines(x)


@wmul_rivendell_cli.command()
@click.argument('source_paths', type=click.Path(exists=True, file_okay=False, dir_okay=True, readable=True), nargs=-1)
@click.option('--cache_duration', type=int, default=180,
              help="How long (in seconds) this importer will remember a given file name after sending it to the "
                   "Rivendell importer. For this duration, this importer will ignore any other files with this name. "
                   "Defaults to 180 seconds (3 minutes).")
@click.option('--rdimport_syslog', is_flag=True, cls=MXWith, mx_with='rdimport_log_file_name',
              help="Tell rdimport to log to syslog.")
@click.option('--rdimport_log_file_name',
              type=click.Path(file_okay=True, dir_okay=False, readable=True, writable=True), cls=MXWith,
              mx_with="rdimport_syslog",  help="Tell rdimport to log to this filename.")
def import_with_file_system_metadata(source_paths, cache_duration, rdimport_syslog, rdimport_log_file_name):
    _logger.debug(f"With {locals()}")
    source_paths = [Path(sp) for sp in source_paths]

    if rdimport_log_file_name:
        rdimport_log_file_name = str(rdimport_log_file_name)
    else:
        rdimport_log_file_name = ""

    args = ImportRivendellFileWithFileSystemMetadataArguments(
        source_paths=source_paths,
        cache_duration=cache_duration,
        rdimport_syslog=rdimport_syslog,
        rdimport_log_file_name=rdimport_log_file_name
    )
    try:
        import_rivendell_file(arguments=args)
    except Exception as e:
        _logger.exception(f"Final crash: {e}")


def get_excluded_groups(excluded_groups_file_name):
    if excluded_groups_file_name:
        with open(excluded_groups_file_name, "rt") as excluded_groups_reader:
            return [excluded_group.strip("\n\r") for excluded_group in excluded_groups_reader]
    else:
        return []




"""
@Author = 'Michael Stanley'

============ Change Log ============
2023-Jan-20 = Change license from GPLv2 to GPLv3.

2020-Oct-08 = Created.

============ License ============
Copyright (C) 2020, 2023 Michael Stanley

This file is part of wmul_rivendell.

wmul_rivendell is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License as published by the Free 
Software Foundation, either version 3 of the License, or (at your option) any 
later version.

wmul_rivendell is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
wmul_rivendell. If not, see <https://www.gnu.org/licenses/>. 
"""
import pytest
import wmul_test_utils

from collections import namedtuple
from wmul_rivendell import RivendellAudioImporter


_for_importer_or_seen_this_time_options = namedtuple(
    "_for_importer_or_seen_this_time_options",
    [
        "new_info_newer"
    ]
)
_for_importer_or_seen_this_time_params, _for_importer_or_seen_this_time_ids = \
    wmul_test_utils.generate_true_false_matrix_from_namedtuple(_for_importer_or_seen_this_time_options)


@pytest.fixture(scope="function", params=_for_importer_or_seen_this_time_params,
                ids=_for_importer_or_seen_this_time_ids)
def setup__for_importer_or_seen_this_time(request, mocker):
    param = request.param

    mock_new_info_much_newer = mocker.Mock(return_value=param.new_info_newer)
    mock_previously_seen_version = mocker.Mock(new_info_much_newer=mock_new_info_much_newer)
    mock_this_version = "mock_this_version"

    result_previous_version, result_this_version = \
        RivendellAudioImporter._for_importer_or_seen_this_time(
            mock_previously_seen_version,
            mock_this_version
        )

    return wmul_test_utils.make_namedtuple(
        "setup__for_importer_or_seen_this_time",
        param=param,
        mock_new_info_much_newer=mock_new_info_much_newer,
        mock_previously_seen_version=mock_previously_seen_version,
        mock_this_version=mock_this_version,
        result_previous_version=result_previous_version,
        result_this_version=result_this_version
    )


def test_new_info_much_newer_called_correctly(setup__for_importer_or_seen_this_time):
    setup__for_importer_or_seen_this_time.mock_new_info_much_newer.assert_called_once_with(
        setup__for_importer_or_seen_this_time.mock_this_version, 30)


def test_correct_file_path_returned(setup__for_importer_or_seen_this_time):
    if setup__for_importer_or_seen_this_time.param.new_info_newer:
        assert setup__for_importer_or_seen_this_time.result_this_version == \
               setup__for_importer_or_seen_this_time.mock_this_version
    else:
        assert setup__for_importer_or_seen_this_time.result_this_version is None


def test_correct_previously_seen_version_returned(setup__for_importer_or_seen_this_time):
    if setup__for_importer_or_seen_this_time.param.new_info_newer:
        assert setup__for_importer_or_seen_this_time.result_previous_version is None
    else:
        assert setup__for_importer_or_seen_this_time.result_previous_version == \
               setup__for_importer_or_seen_this_time.mock_previously_seen_version




"""
@Author = 'Michael Stanley'

============ Change Log ============
2025-Jan-08 = Created. 

============ License ============
Copyright (C) 2025 Michael Stanley

This file is part of wmul_rivendell.

wmul_rivendell is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License as published by the Free 
Software Foundation, either version 3 of the License, or (at your option) any 
later version.

wmul_rivendell is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
wmul_rivendell. If not, see <https://www.gnu.org/licenses/>. 
"""
import pathlib
import pytest
import wmul_test_utils

from wmul_rivendell.LoadCartDataDump import _fix_rivendell_csv_file


@pytest.fixture(scope="function")
def expected_result():
    return \
        'CART_NUMBER,CUT_NUMBER,TYPE,GROUP_NAME,TITLE,ARTIST,ALBUM,YEAR,ISRC,ISCI,LABEL,CLIENT,AGENCY,PUBLISHER,COMPOSER,CONDUCTOR,SONG_ID,USER_DEFINED,DESCRIPTION,OUTCUE,FILENAME,LENGTH,START_POINT,END_POINT,SEGUE_START_POINT,SEGUE_END_POINT,HOOK_START_POINT,HOOK_END_POINT,TALK_START_POINT,TALK_END_POINT,FADEUP_POINT,FADEDOWN_POINT,SCHED_CODES\r\n' \
        '599159,1,audio,aliquid,Inventore neque dignissimos repellendus.,Reprehenderit officiis cum.,,,,,,,,,,,,Laudantium asperiores veritatis quia facere in consequuntur.,Fugit eligendi perferendis itaque placeat optio.,,,,,,,,,,,,,,2010s\r\n' \
        '139437,1,audio,odit,Quisquam voluptatem voluptas nostrum.,Aliquid a consequatur.,,,,,,,,,,,,Modi quo dolores pariatur iste.,Aperiam nesciunt sapiente culpa dolore impedit alias.,,,,,,,,,,,,,,2010s\r\n' \
        '520726,1,audio,incidunt,Repellendus omnis modi sequi.,Corrupti necessitatibus fugiat.,,,,,,,,,,,,Corrupti vitae quo ipsam voluptatem deleniti expedita voluptas.,Doloribus earum nulla debitis blanditiis iste cumque.,,,,,,,,,,,,,,2010s\r\n' \
        '680237,1,audio,doloribus,Quod laboriosam expedita ipsa.,A quae ex.,,,,,,,,,,,,Doloribus ea illo culpa ullam libero.,Qui voluptatibus earum laboriosam libero.,,,,,,,,,,,,,,2010s\r\n' \
        '174387,1,audio,porro,Perferendis tempore consequatur molestiae.,Expedita harum praesentium.,,,,,,,,,,,,"Ipsam voluptas impedit alias a modi amet.",Corrupti non rerum reiciendis.,,,,,,,,,,,,,,2010s\r\n' \
        '146066,1,audio,officia,Similique quo ullam exercitationem.,Ea ratione itaque.,,,,,,,,,,,,Officiis tenetur corrupti dolores nulla porro.,Nemo occaecati ad blanditiis eos sed doloribus enim.,,,,,,,,,,,,,,2010s\r\n' \
        '93372,1,audio,aperiam,Nisi mollitia voluptates autem.,Repellendus cum labore.,,,,,,,,,,,,Nisi ullam deleniti expedita aspernatur deserunt.,Voluptates placeat est vitae sint.,,,,,,,,,,,,,,2010s\r\n' \
        '787912,1,audio,voluptate,Repudiandae quam ipsum quia.,Dignissimos nisi blanditiis.,,,,,,,,,,,,Ab incidunt illum fugit repellat dicta quo reiciendis.,Expedita facere incidunt deserunt.,,,,,,,,,,,,,,2010s\r\n' \
        '785851,1,audio,id,Iure eos libero tenetur.,Illo odit expedita.,,,,,,,,,,,,Laboriosam ipsum labore laboriosam porro.,Reiciendis illo recusandae vitae voluptatibus quam.,,,,,,,,,,,,,,2010s\r\n' \
        '68669,1,audio,saepe,Sed saepe accusamus eveniet.,Eum qui nostrum.,,,,,,,,,,,,Non dolores est nam repellendus.,Dolore quia veritatis.,,,,,,,,,,,,,,2010s\r\n' 

def test_no_extra_new_lines(fs, expected_result):
    source_file_contents = \
        'CART_NUMBER,CUT_NUMBER,TYPE,GROUP_NAME,TITLE,ARTIST,ALBUM,YEAR,ISRC,ISCI,LABEL,CLIENT,AGENCY,PUBLISHER,COMPOSER,CONDUCTOR,SONG_ID,USER_DEFINED,DESCRIPTION,OUTCUE,FILENAME,LENGTH,START_POINT,END_POINT,SEGUE_START_POINT,SEGUE_END_POINT,HOOK_START_POINT,HOOK_END_POINT,TALK_START_POINT,TALK_END_POINT,FADEUP_POINT,FADEDOWN_POINT,SCHED_CODES\r\n' \
        '599159,1,audio,aliquid,Inventore neque dignissimos repellendus.,Reprehenderit officiis cum.,,,,,,,,,,,,Laudantium asperiores veritatis quia facere in consequuntur.,Fugit eligendi perferendis itaque placeat optio.,,,,,,,,,,,,,,2010s\r\n' \
        '139437,1,audio,odit,Quisquam voluptatem voluptas nostrum.,Aliquid a consequatur.,,,,,,,,,,,,Modi quo dolores pariatur iste.,Aperiam nesciunt sapiente culpa dolore impedit alias.,,,,,,,,,,,,,,2010s\r\n' \
        '520726,1,audio,incidunt,Repellendus omnis modi sequi.,Corrupti necessitatibus fugiat.,,,,,,,,,,,,Corrupti vitae quo ipsam voluptatem deleniti expedita voluptas.,Doloribus earum nulla debitis blanditiis iste cumque.,,,,,,,,,,,,,,2010s\r\n' \
        '680237,1,audio,doloribus,Quod laboriosam expedita ipsa.,A quae ex.,,,,,,,,,,,,Doloribus ea illo culpa ullam libero.,Qui voluptatibus earum laboriosam libero.,,,,,,,,,,,,,,2010s\r\n' \
        '174387,1,audio,porro,Perferendis tempore consequatur molestiae.,Expedita harum praesentium.,,,,,,,,,,,,"Ipsam voluptas impedit alias a modi amet.",Corrupti non rerum reiciendis.,,,,,,,,,,,,,,2010s\r\n' \
        '146066,1,audio,officia,Similique quo ullam exercitationem.,Ea ratione itaque.,,,,,,,,,,,,Officiis tenetur corrupti dolores nulla porro.,Nemo occaecati ad blanditiis eos sed doloribus enim.,,,,,,,,,,,,,,2010s\r\n' \
        '93372,1,audio,aperiam,Nisi mollitia voluptates autem.,Repellendus cum labore.,,,,,,,,,,,,Nisi ullam deleniti expedita aspernatur deserunt.,Voluptates placeat est vitae sint.,,,,,,,,,,,,,,2010s\r\n' \
        '787912,1,audio,voluptate,Repudiandae quam ipsum quia.,Dignissimos nisi blanditiis.,,,,,,,,,,,,Ab incidunt illum fugit repellat dicta quo reiciendis.,Expedita facere incidunt deserunt.,,,,,,,,,,,,,,2010s\r\n' \
        '785851,1,audio,id,Iure eos libero tenetur.,Illo odit expedita.,,,,,,,,,,,,Laboriosam ipsum labore laboriosam porro.,Reiciendis illo recusandae vitae voluptatibus quam.,,,,,,,,,,,,,,2010s\r\n' \
        '68669,1,audio,saepe,Sed saepe accusamus eveniet.,Eum qui nostrum.,,,,,,,,,,,,Non dolores est nam repellendus.,Dolore quia veritatis.,,,,,,,,,,,,,,2010s\r\n' 
    
    file_name = pathlib.Path(r"\fakepath\source_file.csv")

    fs.create_file(
        file_name,
        contents=source_file_contents
    )

    with open(file_name, newline="", mode="rt", errors="replace") as rivendell_file:
        result = _fix_rivendell_csv_file(rivendell_source_file=rivendell_file)

    assert expected_result == result.getvalue()


def test_one_with_extra_new_line(fs, expected_result):
    source_file_contents = \
        'CART_NUMBER,CUT_NUMBER,TYPE,GROUP_NAME,TITLE,ARTIST,ALBUM,YEAR,ISRC,ISCI,LABEL,CLIENT,AGENCY,PUBLISHER,COMPOSER,CONDUCTOR,SONG_ID,USER_DEFINED,DESCRIPTION,OUTCUE,FILENAME,LENGTH,START_POINT,END_POINT,SEGUE_START_POINT,SEGUE_END_POINT,HOOK_START_POINT,HOOK_END_POINT,TALK_START_POINT,TALK_END_POINT,FADEUP_POINT,FADEDOWN_POINT,SCHED_CODES\r\n' \
        '599159,1,audio,aliquid,Inventore neque dignissimos repellendus.,Reprehenderit officiis cum.,,,,,,,,,,,,Laudantium asperiores veritatis quia facere in consequuntur.,Fugit eligendi perferendis itaque placeat optio.,,,,,,,,,,,,,,2010s\r\n' \
        '139437,1,audio,odit,Quisquam voluptatem voluptas nostrum.,Aliquid a consequatur.,,,,,,,,,,,,Modi quo dolores pariatur iste.,Aperiam nesciunt sapiente culpa dolore impedit alias.,,,,,,,,,,,,,,2010s\r\n' \
        '520726,1,audio,incidunt,Repellendus omnis modi sequi.,Corrupti necessitatibus fugiat.,,,,,,,,,,,,Corrupti vitae quo ipsam voluptatem deleniti expedita voluptas.,Doloribus earum nulla debitis blanditiis iste cumque.,,,,,,,,,,,,,,2010s\r\n' \
        '680237,1,audio,doloribus,Quod laboriosam expedita ipsa.,A quae ex.,,,,,,,,,,,,Doloribus ea illo culpa ullam libero.,Qui voluptatibus earum laboriosam libero.,,,,,,,,,,,,,,2010s\r\n' \
        '174387,1,audio,porro,Perferendis tempore consequatur molestiae.,Expedita harum praesentium.,,,,,,,,,,,,"Ipsam voluptas impedit\r\nalias a modi amet.",Corrupti non rerum reiciendis.,,,,,,,,,,,,,,2010s\r\n' \
        '146066,1,audio,officia,Similique quo ullam exercitationem.,Ea ratione itaque.,,,,,,,,,,,,Officiis tenetur corrupti dolores nulla porro.,Nemo occaecati ad blanditiis eos sed doloribus enim.,,,,,,,,,,,,,,2010s\r\n' \
        '93372,1,audio,aperiam,Nisi mollitia voluptates autem.,Repellendus cum labore.,,,,,,,,,,,,Nisi ullam deleniti expedita aspernatur deserunt.,Voluptates placeat est vitae sint.,,,,,,,,,,,,,,2010s\r\n' \
        '787912,1,audio,voluptate,Repudiandae quam ipsum quia.,Dignissimos nisi blanditiis.,,,,,,,,,,,,Ab incidunt illum fugit repellat dicta quo reiciendis.,Expedita facere incidunt deserunt.,,,,,,,,,,,,,,2010s\r\n' \
        '785851,1,audio,id,Iure eos libero tenetur.,Illo odit expedita.,,,,,,,,,,,,Laboriosam ipsum labore laboriosam porro.,Reiciendis illo recusandae vitae voluptatibus quam.,,,,,,,,,,,,,,2010s\r\n' \
        '68669,1,audio,saepe,Sed saepe accusamus eveniet.,Eum qui nostrum.,,,,,,,,,,,,Non dolores est nam repellendus.,Dolore quia veritatis.,,,,,,,,,,,,,,2010s\r\n' 
         
    file_name = pathlib.Path(r"\fakepath\source_file.csv")

    fs.create_file(
        file_name,
        contents=source_file_contents
    )

    with open(file_name, newline="", mode="rt", errors="replace") as rivendell_file:
        result = _fix_rivendell_csv_file(rivendell_source_file=rivendell_file)

    assert expected_result == result.getvalue()

def test_two_with_extra_new_line(fs, expected_result):
    source_file_contents = \
        'CART_NUMBER,CUT_NUMBER,TYPE,GROUP_NAME,TITLE,ARTIST,ALBUM,YEAR,ISRC,ISCI,LABEL,CLIENT,AGENCY,PUBLISHER,COMPOSER,CONDUCTOR,SONG_ID,USER_DEFINED,DESCRIPTION,OUTCUE,FILENAME,LENGTH,START_POINT,END_POINT,SEGUE_START_POINT,SEGUE_END_POINT,HOOK_START_POINT,HOOK_END_POINT,TALK_START_POINT,TALK_END_POINT,FADEUP_POINT,FADEDOWN_POINT,SCHED_CODES\r\n' \
        '599159,1,audio,aliquid,Inventore neque dignissimos repellendus.,Reprehenderit officiis cum.,,,,,,,,,,,,Laudantium asperiores veritatis quia facere in consequuntur.,Fugit eligendi perferendis itaque placeat optio.,,,,,,,,,,,,,,2010s\r\n' \
        '139437,1,audio,odit,Quisquam voluptatem voluptas nostrum.,Aliquid a consequatur.,,,,,,,,,,,,Modi quo dolores pariatur iste.,Aperiam nesciunt sapiente culpa dolore impedit alias.,,,,,,,,,,,,,,2010s\r\n' \
        '520726,1,audio,incidunt,Repellendus omnis modi sequi.,Corrupti necessitatibus fugiat.,,,,,,,,,,,,Corrupti vitae quo ipsam voluptatem deleniti expedita voluptas.,Doloribus earum nulla debitis blanditiis iste cumque.,,,,,,,,,,,,,,2010s\r\n' \
        '680237,1,audio,doloribus,Quod laboriosam expedita ipsa.,A quae ex.,,,,,,,,,,,,Doloribus ea illo culpa ullam libero.,Qui voluptatibus earum laboriosam libero.,,,,,,,,,,,,,,2010s\r\n' \
        '174387,1,audio,porro,Perferendis tempore consequatur molestiae.,Expedita harum praesentium.,,,,,,,,,,,,"Ipsam voluptas impedit\r\nalias a modi amet.",Corrupti non rerum reiciendis.,,,,,,,,,,,,,,2010s\r\n' \
        '146066,1,audio,officia,Similique quo ullam exercitationem.,Ea ratione itaque.,,,,,,,,,,,,Officiis tenetur corrupti\r\ndolores nulla porro.,Nemo occaecati ad blanditiis eos sed doloribus enim.,,,,,,,,,,,,,,2010s\r\n' \
        '93372,1,audio,aperiam,Nisi mollitia voluptates autem.,Repellendus cum labore.,,,,,,,,,,,,Nisi ullam deleniti expedita aspernatur deserunt.,Voluptates placeat est vitae sint.,,,,,,,,,,,,,,2010s\r\n' \
        '787912,1,audio,voluptate,Repudiandae quam ipsum quia.,Dignissimos nisi blanditiis.,,,,,,,,,,,,Ab incidunt illum fugit repellat dicta quo reiciendis.,Expedita facere incidunt deserunt.,,,,,,,,,,,,,,2010s\r\n' \
        '785851,1,audio,id,Iure eos libero tenetur.,Illo odit expedita.,,,,,,,,,,,,Laboriosam ipsum labore laboriosam porro.,Reiciendis illo recusandae vitae voluptatibus quam.,,,,,,,,,,,,,,2010s\r\n' \
        '68669,1,audio,saepe,Sed saepe accusamus eveniet.,Eum qui nostrum.,,,,,,,,,,,,Non dolores est nam repellendus.,Dolore quia veritatis.,,,,,,,,,,,,,,2010s\r\n' 

    file_name = pathlib.Path(r"\fakepath\source_file.csv")

    fs.create_file(
        file_name,
        contents=source_file_contents
    )

    with open(file_name, newline="", mode="rt", errors="replace") as rivendell_file:
        result = _fix_rivendell_csv_file(rivendell_source_file=rivendell_file)

    assert expected_result == result.getvalue()


def test_multiple_extra_new_line(fs, expected_result):
    source_file_contents = \
        'CART_NUMBER,CUT_NUMBER,TYPE,GROUP_NAME,TITLE,ARTIST,ALBUM,YEAR,ISRC,ISCI,LABEL,CLIENT,AGENCY,PUBLISHER,COMPOSER,CONDUCTOR,SONG_ID,USER_DEFINED,DESCRIPTION,OUTCUE,FILENAME,LENGTH,START_POINT,END_POINT,SEGUE_START_POINT,SEGUE_END_POINT,HOOK_START_POINT,HOOK_END_POINT,TALK_START_POINT,TALK_END_POINT,FADEUP_POINT,FADEDOWN_POINT,SCHED_CODES\r\n' \
        '599159,1,audio,aliquid,Inventore neque dignissimos repellendus.,Reprehenderit officiis cum.,,,,,,,,,,,,Laudantium asperiores veritatis quia facere in consequuntur.,Fugit eligendi perferendis itaque placeat optio.,,,,,,,,,,,,,,2010s\r\n' \
        '139437,1,audio,odit,Quisquam voluptatem voluptas nostrum.,Aliquid a consequatur.,,,,,,,,,,,,Modi quo dolores pariatur iste.,Aperiam nesciunt sapiente culpa dolore impedit alias.,,,,,,,,,,,,,,2010s\r\n' \
        '520726,1,audio,incidunt,Repellendus omnis modi sequi.,Corrupti necessitatibus fugiat.,,,,,,,,,,,,Corrupti vitae quo ipsam voluptatem deleniti expedita voluptas.,Doloribus earum nulla debitis blanditiis iste cumque.,,,,,,,,,,,,,,2010s\r\n' \
        '680237,1,audio,doloribus,Quod laboriosam expedita ipsa.,A quae ex.,,,,,,,,,,,,Doloribus ea illo culpa ullam libero.,Qui voluptatibus earum laboriosam libero.,,,,,,,,,,,,,,2010s\r\n' \
        '174387,1,audio,porro,Perferendis tempore consequatur molestiae.,Expedita harum praesentium.,,,,,,,,,,,,"Ipsam voluptas impedit\r\nalias a modi amet.",Corrupti non rerum reiciendis.,,,,,,,,,,,,,,2010s\r\n' \
        '146066,1,audio,officia,Similique\r\nquo ullam exercitationem.,Ea ratione itaque.,,,,,,,,,,,,Officiis tenetur corrupti\r\ndolores nulla porro.,Nemo occaecati ad blanditiis eos sed doloribus enim.,,,,,,,,,,,,,,2010s\r\n' \
        '93372,1,audio,aperiam,Nisi mollitia voluptates autem.,Repellendus cum labore.,,,,,,,,,,,,Nisi ullam deleniti expedita aspernatur deserunt.,Voluptates placeat est vitae sint.,,,,,,,,,,,,,,2010s\r\n' \
        '787912,1,audio,voluptate,Repudiandae quam ipsum quia.,Dignissimos nisi blanditiis.,,,,,,,,,,,,Ab incidunt illum fugit repellat dicta quo reiciendis.,Expedita facere incidunt deserunt.,,,,,,,,,,,,,,2010s\r\n' \
        '785851,1,audio,id,Iure eos libero tenetur.,Illo odit expedita.,,,,,,,,,,,,Laboriosam ipsum labore laboriosam porro.,Reiciendis illo recusandae vitae voluptatibus quam.,,,,,,,,,,,,,,2010s\r\n' \
        '68669,1,audio,saepe,Sed saepe accusamus eveniet.,Eum qui nostrum.,,,,,,,,,,,,Non dolores est nam repellendus.,Dolore quia veritatis.,,,,,,,,,,,,,,2010s\r\n' 
    
    file_name = pathlib.Path(r"\fakepath\source_file.csv")

    fs.create_file(
        file_name,
        contents=source_file_contents
    )

    with open(file_name, newline="", mode="rt", errors="replace") as rivendell_file:
        result = _fix_rivendell_csv_file(rivendell_source_file=rivendell_file)

    assert expected_result == result.getvalue()


"""
@Author = 'Michael Stanley'

============ Change Log ============
2023-May-25 = Created. Most of this module was refactored from 
                tests/FilterCartReportForMusicScheduler/
                test_filtercartreportformusicscheduler.py when the code under 
                test was refactored into LoadCartDataDump.

============ License ============
Copyright (C) 2021, 2023, 2025 Michael Stanley

This file is part of wmul_rivendell.

wmul_rivendell is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License as published by the Free 
Software Foundation, either version 3 of the License, or (at your option) any 
later version.

wmul_rivendell is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
wmul_rivendell. If not, see <https://www.gnu.org/licenses/>. 
"""
import pytest
import wmul_test_utils

from wmul_rivendell.LoadCartDataDump import LoadCartDataDump, RivendellCart, CartType

@pytest.fixture(scope="function")
def setup_standard_cart_filter(cart_source_file_contents):
    mock_rivendell_cart_data_filename = "mock_rivendell_cart_data_filename"

    cart_filter = LoadCartDataDump(
        rivendell_cart_data_filename=mock_rivendell_cart_data_filename,
        include_macros=False,
        excluded_group_list=[],
        include_all_cuts=False
    )

    return wmul_test_utils.make_namedtuple(
        "setup_standard_cart_filter",
        cart_filter=cart_filter,
        cart_source_file_contents=\
            cart_source_file_contents.source_file_contents,
    )


def test__load_rivendell_carts_file_does_not_exist(
        fs, 
        setup_standard_cart_filter
    ):

    cart_filter = setup_standard_cart_filter.cart_filter
    import pathlib
    rivendell_cart_data_filename = pathlib.Path(r"\fakepath\fakefile.csv")
    assert not rivendell_cart_data_filename.exists()

    output_file_path = pathlib.Path(r"\fakepath\fake_output_file.csv")

    cart_filter.rivendell_cart_data_filename = rivendell_cart_data_filename
    cart_filter.output_filename = output_file_path

    with pytest.raises(OSError) as oserror:
        cart_filter._load_rivendell_carts()


@pytest.fixture(scope="function")
def cart_source_file_contents():
    """Source file contents should parse to rivendell_cart_1_1, rivendell_cart_5_1, rivendell_cart_5071_1, and 
       rivendell_cart_100249_1 in the defined_carts fixture."""

    source_file_contents = \
        'CART_NUMBER,CUT_NUMBER,TYPE,GROUP_NAME,TITLE,ARTIST,ALBUM,YEAR,ISRC,ISCI,LABEL,CLIENT,AGENCY,PUBLISHER,COMPOSER,CONDUCTOR,SONG_ID,USER_DEFINED,DESCRIPTION,OUTCUE,FILENAME,LENGTH,START_POINT,END_POINT,SEGUE_START_POINT,SEGUE_END_POINT,HOOK_START_POINT,HOOK_END_POINT,TALK_START_POINT,TALK_END_POINT,FADEUP_POINT,FADEDOWN_POINT,SCHED_CODES\r\n' \
        '1,1,audio,TEMPORIBUS,perferendis optio adipisci odit,nesciunt repellendus,,1951,,,,saepe architecto sequi alias,,,facere odio magnam odit,,,,perferendis pariatur eligendi dolorem,,000001_001.wav,:07,,,,,,,,,,,2010s\r\n' \
        '2,1,audio,TEMPORIBUS,saepe recusandae deleniti laborum,quisquam natus,,1965,,,,enim ut qui veniam,,,omnis voluptatum accusantium saepe,,,,praesentium eaque deleniti tempore,,000002_001.wav,:07,,,,,,,,,,,2010s\r\n' \
        '2,2,audio,TEMPORIBUS,aperiam sed totam reiciendis,recusandae nesciunt,,1962,,,,fugiat saepe quam libero,,,repellat non in fuga,,,,deserunt repellat odit velit,,000002_002.wav,:11,,,,,,,,,,,2010s\r\n' \
        '6,2,audio,TEMPORIBUS,tempora libero earum repudiandae,modi quo,,1953,,,,iure nobis illum quia,,,fugiat beatae perspiciatis dicta,,,,tempore voluptates tenetur dolorem,,000006_002.wav,:11,,,,,,,,,,,2010s\r\n' \
        '101,1,audio,TEMPORIBUS,aliquam omnis quisquam enim,nobis sit,,2003,,,,culpa quas minus reiciendis,,,ipsam cumque nulla harum,,,,deserunt natus quidem at,,000101_001.wav,:05,,,,,,,,,,,2010s\r\n' \
        '500,1,audio,VOLUPTATIB,voluptate commodi neque aperiam,reiciendis rem,,1969,,,,assumenda quasi maxime cum,,,doloremque provident dolorum mollitia,,,,ut praesentium sit quidem,,000500_001.wav,:08,,,,,,,,,,,2010s\r\n' \
        '970000,1,macro,ALIQUAM,quasi necessitatibus quo vero,sapiente fuga,,1994,,,,distinctio culpa perspiciatis quam,,,explicabo magni temporibus provident,,,,adipisci officia impedit dolores,,,:00,,,,,,,,,,,2010s\r\n' \
        '970001,1,macro,ALIQUAM,porro eos dolores fuga,rem dolor,,2015,,,,omnis molestias nesciunt cumque,,,adipisci quidem consequuntur quia,,,,provident perspiciatis quisquam ea,,,:00,,,,,,,,,,,2010s\r\n'
    
    return wmul_test_utils.make_namedtuple(
        "cart_source_file_contents",
        source_file_contents=source_file_contents
    )


@pytest.fixture(scope="function")
def defined_rivendell_carts():
    rivendell_cart_1_1 = RivendellCart(cart_number='1', cut_number='1', type=CartType.Audio, group_name='TEMPORIBUS', title='perferendis optio adipisci odit', artist='nesciunt repellendus', album='', year='1951', isrc='', isci='', label='', client='saepe architecto sequi alias', agency='', publisher='', composer='facere odio magnam odit', conductor='', song_id='', user_defined='', description='perferendis pariatur eligendi dolorem', outcue='', filename='000001_001.wav', length=':07', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes='2010s')
    rivendell_cart_2_1 = RivendellCart(cart_number='2', cut_number='1', type=CartType.Audio, group_name='TEMPORIBUS', title='saepe recusandae deleniti laborum', artist='quisquam natus', album='', year='1965', isrc='', isci='', label='', client='enim ut qui veniam', agency='', publisher='', composer='omnis voluptatum accusantium saepe', conductor='', song_id='', user_defined='', description='praesentium eaque deleniti tempore', outcue='', filename='000002_001.wav', length=':07', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes='2010s')
    rivendell_cart_2_2 = RivendellCart(cart_number='2', cut_number='2', type=CartType.Audio, group_name='TEMPORIBUS', title='aperiam sed totam reiciendis', artist='recusandae nesciunt', album='', year='1962', isrc='', isci='', label='', client='fugiat saepe quam libero', agency='', publisher='', composer='repellat non in fuga', conductor='', song_id='', user_defined='', description='deserunt repellat odit velit', outcue='', filename='000002_002.wav', length=':11', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes='2010s')
    rivendell_cart_6_2 = RivendellCart(cart_number='6', cut_number='2', type=CartType.Audio, group_name='TEMPORIBUS', title='tempora libero earum repudiandae', artist='modi quo', album='', year='1953', isrc='', isci='', label='', client='iure nobis illum quia', agency='', publisher='', composer='fugiat beatae perspiciatis dicta', conductor='', song_id='', user_defined='', description='tempore voluptates tenetur dolorem', outcue='', filename='000006_002.wav', length=':11', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes='2010s')
    rivendell_cart_101_1 = RivendellCart(cart_number='101', cut_number='1', type=CartType.Audio, group_name='TEMPORIBUS', title='aliquam omnis quisquam enim', artist='nobis sit', album='', year='2003', isrc='', isci='', label='', client='culpa quas minus reiciendis', agency='', publisher='', composer='ipsam cumque nulla harum', conductor='', song_id='', user_defined='', description='deserunt natus quidem at', outcue='', filename='000101_001.wav', length=':05', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes='2010s')
    rivendell_cart_500_1 = RivendellCart(cart_number='500', cut_number='1', type=CartType.Audio, group_name='VOLUPTATIB', title='voluptate commodi neque aperiam', artist='reiciendis rem', album='', year='1969', isrc='', isci='', label='', client='assumenda quasi maxime cum', agency='', publisher='', composer='doloremque provident dolorum mollitia', conductor='', song_id='', user_defined='', description='ut praesentium sit quidem', outcue='', filename='000500_001.wav', length=':08', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes='2010s')
    rivendell_cart_970000_1 = RivendellCart(cart_number='970000', cut_number='1', type=CartType.Macro, group_name='ALIQUAM', title='quasi necessitatibus quo vero', artist='sapiente fuga', album='', year='1994', isrc='', isci='', label='', client='distinctio culpa perspiciatis quam', agency='', publisher='', composer='explicabo magni temporibus provident', conductor='', song_id='', user_defined='', description='adipisci officia impedit dolores', outcue='', filename='', length=':00', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes='2010s')
    rivendell_cart_970001_1 = RivendellCart(cart_number='970001', cut_number='1', type=CartType.Macro, group_name='ALIQUAM', title='porro eos dolores fuga', artist='rem dolor', album='', year='2015', isrc='', isci='', label='', client='omnis molestias nesciunt cumque', agency='', publisher='', composer='adipisci quidem consequuntur quia', conductor='', song_id='', user_defined='', description='provident perspiciatis quisquam ea', outcue='', filename='', length=':00', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes='2010s')
    return wmul_test_utils.make_namedtuple(
        "defined_carts",
        rivendell_cart_1_1=rivendell_cart_1_1,
        rivendell_cart_2_1=rivendell_cart_2_1,
        rivendell_cart_2_2=rivendell_cart_2_2,
        rivendell_cart_6_2=rivendell_cart_6_2,
        rivendell_cart_101_1=rivendell_cart_101_1,
        rivendell_cart_500_1=rivendell_cart_500_1,
        rivendell_cart_970000_1=rivendell_cart_970000_1,
        rivendell_cart_970001_1=rivendell_cart_970001_1
    )


def test__load_rivendell_carts(fs, setup_standard_cart_filter, 
                               defined_rivendell_carts):
    cart_filter = setup_standard_cart_filter.cart_filter
    import pathlib
    rivendell_cart_data_filename = pathlib.Path(r"\fakepath\source_file.csv")
    assert not rivendell_cart_data_filename.exists()

    output_file_path = pathlib.Path(r"\fakepath\output_file.csv")

    fs.create_file(
        rivendell_cart_data_filename, 
        contents=setup_standard_cart_filter.cart_source_file_contents
    )

    cart_filter.rivendell_cart_data_filename = rivendell_cart_data_filename
    cart_filter.output_filename = output_file_path

    result_carts = cart_filter._load_rivendell_carts()

    expected_carts = [
        defined_rivendell_carts.rivendell_cart_1_1,
        defined_rivendell_carts.rivendell_cart_2_1,
        defined_rivendell_carts.rivendell_cart_2_2,
        defined_rivendell_carts.rivendell_cart_6_2,
        defined_rivendell_carts.rivendell_cart_101_1,
        defined_rivendell_carts.rivendell_cart_500_1,
        defined_rivendell_carts.rivendell_cart_970000_1,
        defined_rivendell_carts.rivendell_cart_970001_1
    ]
    x = 1
    assert result_carts == expected_carts


def test__remove_macro_carts_no_carts(setup_standard_cart_filter):
    rivendell_carts = []
    expected_carts = []
    result_carts = setup_standard_cart_filter.cart_filter\
        ._remove_macro_carts(rivendell_carts=rivendell_carts)
    assert list(result_carts) == expected_carts


def test__remove_macro_carts_all_audio_carts(setup_standard_cart_filter, 
                                             defined_rivendell_carts):
    rivendell_carts = [
        defined_rivendell_carts.rivendell_cart_1_1,
        defined_rivendell_carts.rivendell_cart_2_1,
        defined_rivendell_carts.rivendell_cart_2_2,
        defined_rivendell_carts.rivendell_cart_6_2,
        defined_rivendell_carts.rivendell_cart_101_1,
        defined_rivendell_carts.rivendell_cart_500_1
    ]

    expected_carts = [
        defined_rivendell_carts.rivendell_cart_1_1,
        defined_rivendell_carts.rivendell_cart_2_1,
        defined_rivendell_carts.rivendell_cart_2_2,
        defined_rivendell_carts.rivendell_cart_6_2,
        defined_rivendell_carts.rivendell_cart_101_1,
        defined_rivendell_carts.rivendell_cart_500_1
    ]

    result_carts = setup_standard_cart_filter.cart_filter\
        ._remove_macro_carts(rivendell_carts=rivendell_carts)
    assert expected_carts == list(result_carts)


def test__remove_macro_carts_mixed_carts(setup_standard_cart_filter, 
                                         defined_rivendell_carts):
    rivendell_carts = [
        defined_rivendell_carts.rivendell_cart_1_1,
        defined_rivendell_carts.rivendell_cart_2_1,
        defined_rivendell_carts.rivendell_cart_2_2,
        defined_rivendell_carts.rivendell_cart_6_2,
        defined_rivendell_carts.rivendell_cart_101_1,
        defined_rivendell_carts.rivendell_cart_500_1,
        defined_rivendell_carts.rivendell_cart_970000_1,
        defined_rivendell_carts.rivendell_cart_970001_1
    ]

    expected_carts = [
        defined_rivendell_carts.rivendell_cart_1_1,
        defined_rivendell_carts.rivendell_cart_2_1,
        defined_rivendell_carts.rivendell_cart_2_2,
        defined_rivendell_carts.rivendell_cart_6_2,
        defined_rivendell_carts.rivendell_cart_101_1,
        defined_rivendell_carts.rivendell_cart_500_1
    ]

    result_carts = setup_standard_cart_filter.cart_filter\
        ._remove_macro_carts(rivendell_carts=rivendell_carts)
    assert expected_carts == list(result_carts)


def test__remove_macro_carts_all_macro_carts(setup_standard_cart_filter, 
                                             defined_rivendell_carts):
    rivendell_carts = [
        defined_rivendell_carts.rivendell_cart_970000_1,
        defined_rivendell_carts.rivendell_cart_970001_1
    ]
    expected_carts = []

    result_carts = setup_standard_cart_filter.cart_filter.\
        _remove_macro_carts(rivendell_carts=rivendell_carts)
    assert expected_carts == list(result_carts)


def test__remove_extra_cuts_no_cuts(setup_standard_cart_filter):
    rivendell_carts = []
    result_carts = setup_standard_cart_filter.cart_filter.\
        _remove_extra_cuts(rivendell_carts=rivendell_carts)

    assert len(result_carts) == 0
    assert result_carts is not rivendell_carts


def test__remove_extra_cuts_no_extra_cuts(setup_standard_cart_filter, defined_rivendell_carts):
    rivendell_carts_for_test = [
        defined_rivendell_carts.rivendell_cart_1_1,
        defined_rivendell_carts.rivendell_cart_2_1,
        defined_rivendell_carts.rivendell_cart_6_2,
        defined_rivendell_carts.rivendell_cart_101_1,
    ]

    expected_carts = [
        defined_rivendell_carts.rivendell_cart_1_1,
        defined_rivendell_carts.rivendell_cart_2_1,
        defined_rivendell_carts.rivendell_cart_6_2,
        defined_rivendell_carts.rivendell_cart_101_1,
    ]

    result_carts = setup_standard_cart_filter.cart_filter\
        ._remove_extra_cuts(rivendell_carts=rivendell_carts_for_test)

    assert list(result_carts) == expected_carts


def test__remove_extra_cuts_extra_cuts_in_order(setup_standard_cart_filter, 
                                                defined_rivendell_carts):
    rivendell_carts_for_test = [
        defined_rivendell_carts.rivendell_cart_1_1,
        defined_rivendell_carts.rivendell_cart_2_1,
        defined_rivendell_carts.rivendell_cart_2_2,
        defined_rivendell_carts.rivendell_cart_6_2,
        defined_rivendell_carts.rivendell_cart_101_1,
    ]

    expected_carts = [
        defined_rivendell_carts.rivendell_cart_1_1,
        defined_rivendell_carts.rivendell_cart_2_1,
        defined_rivendell_carts.rivendell_cart_6_2,
        defined_rivendell_carts.rivendell_cart_101_1,
    ]

    result_carts = setup_standard_cart_filter.cart_filter\
        ._remove_extra_cuts(rivendell_carts=rivendell_carts_for_test)

    assert list(result_carts) == expected_carts


def test__remove_extra_cuts_extra_cuts_out_of_order(setup_standard_cart_filter, 
                                                    defined_rivendell_carts):
    rivendell_carts_for_test = [
        defined_rivendell_carts.rivendell_cart_1_1,
        defined_rivendell_carts.rivendell_cart_2_2,
        defined_rivendell_carts.rivendell_cart_2_1,
        defined_rivendell_carts.rivendell_cart_6_2,
        defined_rivendell_carts.rivendell_cart_101_1,
    ]

    expected_carts = [
        defined_rivendell_carts.rivendell_cart_1_1,
        defined_rivendell_carts.rivendell_cart_2_1,
        defined_rivendell_carts.rivendell_cart_6_2,
        defined_rivendell_carts.rivendell_cart_101_1,
    ]

    result_carts = setup_standard_cart_filter.cart_filter\
        ._remove_extra_cuts(rivendell_carts=rivendell_carts_for_test)

    assert list(result_carts) == expected_carts


def test__remove_excluded_groups_empty_list(setup_standard_cart_filter, defined_rivendell_carts):
    rivendell_carts_for_test = [
        defined_rivendell_carts.rivendell_cart_1_1,
        defined_rivendell_carts.rivendell_cart_2_1,
        defined_rivendell_carts.rivendell_cart_2_2,
        defined_rivendell_carts.rivendell_cart_6_2,
        defined_rivendell_carts.rivendell_cart_101_1,
        defined_rivendell_carts.rivendell_cart_500_1,
    ]

    expected_carts = [
        defined_rivendell_carts.rivendell_cart_1_1,
        defined_rivendell_carts.rivendell_cart_2_1,
        defined_rivendell_carts.rivendell_cart_2_2,
        defined_rivendell_carts.rivendell_cart_6_2,
        defined_rivendell_carts.rivendell_cart_101_1,
        defined_rivendell_carts.rivendell_cart_500_1,
    ]

    result_carts = setup_standard_cart_filter.cart_filter\
        ._remove_excluded_groups(rivendell_carts=rivendell_carts_for_test)

    assert list(result_carts) == expected_carts


def test__remove_excluded_groups_no_carts_from_groups(setup_standard_cart_filter, defined_rivendell_carts):
    rivendell_carts_for_test = [
        defined_rivendell_carts.rivendell_cart_1_1,
        defined_rivendell_carts.rivendell_cart_2_1,
        defined_rivendell_carts.rivendell_cart_2_2,
        defined_rivendell_carts.rivendell_cart_6_2,
        defined_rivendell_carts.rivendell_cart_101_1,
        defined_rivendell_carts.rivendell_cart_500_1,
    ]

    expected_carts = [
        defined_rivendell_carts.rivendell_cart_1_1,
        defined_rivendell_carts.rivendell_cart_2_1,
        defined_rivendell_carts.rivendell_cart_2_2,
        defined_rivendell_carts.rivendell_cart_6_2,
        defined_rivendell_carts.rivendell_cart_101_1,
        defined_rivendell_carts.rivendell_cart_500_1,
    ]

    excluded_groups = ["POSSIMUS", "INCIDUNT"]
    setup_standard_cart_filter.cart_filter.excluded_group_list = excluded_groups

    result_carts = setup_standard_cart_filter.cart_filter\
        ._remove_excluded_groups(rivendell_carts=rivendell_carts_for_test)

    assert list(result_carts) == expected_carts


def test__remove_excluded_groups_some_carts_from_groups(setup_standard_cart_filter, defined_rivendell_carts):
    rivendell_carts_for_test = [
        defined_rivendell_carts.rivendell_cart_1_1,
        defined_rivendell_carts.rivendell_cart_2_1,
        defined_rivendell_carts.rivendell_cart_2_2,
        defined_rivendell_carts.rivendell_cart_6_2,
        defined_rivendell_carts.rivendell_cart_101_1,
        defined_rivendell_carts.rivendell_cart_500_1,
    ]

    expected_carts = [
        defined_rivendell_carts.rivendell_cart_1_1,
        defined_rivendell_carts.rivendell_cart_2_1,
        defined_rivendell_carts.rivendell_cart_2_2,
        defined_rivendell_carts.rivendell_cart_6_2,
        defined_rivendell_carts.rivendell_cart_101_1,
    ]

    excluded_groups = ["POSSIMUS", "INCIDUNT", "VOLUPTATIB"]
    setup_standard_cart_filter.cart_filter.excluded_group_list = excluded_groups

    result_carts = setup_standard_cart_filter.cart_filter\
        ._remove_excluded_groups(rivendell_carts=rivendell_carts_for_test)

    assert list(result_carts) == expected_carts


def test__remove_excluded_groups_all_carts_from_groups(setup_standard_cart_filter, defined_rivendell_carts):
    rivendell_carts_for_test = [
        defined_rivendell_carts.rivendell_cart_1_1,
        defined_rivendell_carts.rivendell_cart_2_1,
        defined_rivendell_carts.rivendell_cart_2_2,
        defined_rivendell_carts.rivendell_cart_6_2,
        defined_rivendell_carts.rivendell_cart_101_1,
        defined_rivendell_carts.rivendell_cart_500_1,
    ]

    expected_carts = [
    ]

    excluded_groups = ["POSSIMUS", "INCIDUNT", "VOLUPTATIB", "TEMPORIBUS", "EXPLICABO"]
    setup_standard_cart_filter.cart_filter.excluded_group_list = excluded_groups

    result_carts = setup_standard_cart_filter.cart_filter\
        ._remove_excluded_groups(rivendell_carts=rivendell_carts_for_test)

    assert list(result_carts) == expected_carts


load_carts_params, rload_carts_ids = wmul_test_utils\
    .generate_true_false_matrix_from_list_of_strings(
        "load_carts_options",
        [
            "include_macros",
            "include_all_cuts",
            "exclude_groups"
        ]
    )


@pytest.fixture(scope="function", params=load_carts_params, ids=rload_carts_ids)
def setup_run_script(request, caplog, mocker):
    params = request.param
    mock_rivendell_cart_data_filename = "mock_rivendell_cart_data_filename"

    mock_rivendell_carts = "mock_rivendell_carts"
    mock_load_rivendell_carts = mocker.Mock(return_value=mock_rivendell_carts)
    mocker.patch(
        "wmul_rivendell.LoadCartDataDump.LoadCartDataDump._load_rivendell_carts",
        mock_load_rivendell_carts
    )

    mock_rivendell_carts_without_excluded_groups = "mock_rivendell_carts_without_excluded_groups"
    mock_remove_excluded_groups = mocker.Mock(return_value=mock_rivendell_carts_without_excluded_groups)
    mocker.patch(
        "wmul_rivendell.LoadCartDataDump.LoadCartDataDump._remove_excluded_groups",
        mock_remove_excluded_groups
    )

    mock_rivendell_carts_without_macros = "mock_rivendell_carts_without_macros"
    mock_remove_macro_carts = mocker.Mock(return_value=mock_rivendell_carts_without_macros)
    mocker.patch(
        "wmul_rivendell.LoadCartDataDump.LoadCartDataDump._remove_macro_carts",
        mock_remove_macro_carts
    )

    mock_rivendell_carts_without_extra_cuts = "mock_rivendell_carts_without_extra_cuts"
    mock_remove_extra_cuts = mocker.Mock(return_value=mock_rivendell_carts_without_extra_cuts)
    mocker.patch(
        "wmul_rivendell.LoadCartDataDump.LoadCartDataDump._remove_extra_cuts",
        mock_remove_extra_cuts
    )

    cart_filter = LoadCartDataDump(
        rivendell_cart_data_filename=mock_rivendell_cart_data_filename,
        include_macros=params.include_macros,
        excluded_group_list=params.exclude_groups,
        include_all_cuts=params.include_all_cuts
    )

    results = cart_filter.load_carts()

    return wmul_test_utils.make_namedtuple(
        "setup_load_carts",
        params=params,
        mock_load_rivendell_carts=mock_load_rivendell_carts,
        mock_rivendell_carts=mock_rivendell_carts,
        mock_rivendell_carts_without_excluded_groups=mock_rivendell_carts_without_excluded_groups,
        mock_remove_excluded_groups=mock_remove_excluded_groups,
        mock_remove_macro_carts=mock_remove_macro_carts,
        mock_rivendell_carts_without_macros=mock_rivendell_carts_without_macros,
        mock_remove_extra_cuts=mock_remove_extra_cuts,
        mock_rivendell_carts_without_extra_cuts=mock_rivendell_carts_without_extra_cuts,
        results=results,
        caplog_text=caplog.text
    )


def test_run_script_load_rivendell_carts_called_correctly(setup_run_script):
    setup_run_script.mock_load_rivendell_carts.assert_called_once_with()


def test_run_script_remove_macro_carts_called_correctly(setup_run_script):
    if setup_run_script.params.include_macros:
        setup_run_script.mock_remove_macro_carts.assert_not_called()
    else:
        if setup_run_script.params.exclude_groups:
            setup_run_script.mock_remove_macro_carts.assert_called_once_with(setup_run_script.mock_rivendell_carts_without_excluded_groups)
        else:
            setup_run_script.mock_remove_macro_carts.assert_called_once_with(setup_run_script.mock_rivendell_carts)


def test_run_script_remove_extra_cuts_called_correctly(setup_run_script):
    if setup_run_script.params.include_all_cuts:
        setup_run_script.mock_remove_extra_cuts.assert_not_called()
    else:
        if setup_run_script.params.include_macros:
            if setup_run_script.params.exclude_groups:
                setup_run_script.mock_remove_extra_cuts.assert_called_once_with(setup_run_script.mock_rivendell_carts_without_excluded_groups)
                pass
            else:
                setup_run_script.mock_remove_extra_cuts.assert_called_once_with(setup_run_script.mock_rivendell_carts)
        else:
            setup_run_script.mock_remove_extra_cuts.assert_called_once_with(
                setup_run_script.mock_rivendell_carts_without_macros
            )


def test_run_script_returns_correct_item(setup_run_script):
    if not setup_run_script.params.include_all_cuts:
        assert setup_run_script.results == setup_run_script.mock_rivendell_carts_without_extra_cuts
    elif not setup_run_script.params.include_macros:
        assert setup_run_script.results == setup_run_script.mock_rivendell_carts_without_macros
    elif setup_run_script.params.exclude_groups:
        assert setup_run_script.results == setup_run_script.mock_rivendell_carts_without_excluded_groups
    else:
        assert setup_run_script.results == setup_run_script.mock_rivendell_carts



"""
@Author = 'Michael Stanley'

============ Change Log ============
2023-Jan-20 = Change license from GPLv2 to GPLv3.

2023-Jan-11 = Added tests that deal with a bug introduced in version 3.6.4. 
              The bug causes the header for the "FILENAME" and "LENGTH" to be 
              combined. Only the headers are combined, the actual data is 
              properly separated.

2021-Jan-29 = Created.

============ License ============
Copyright (C) 2021, 2023, 2025 Michael Stanley

This file is part of wmul_rivendell.

wmul_rivendell is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License as published by the Free 
Software Foundation, either version 3 of the License, or (at your option) any 
later version.

wmul_rivendell is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
wmul_rivendell. If not, see <https://www.gnu.org/licenses/>. 
"""
from collections import OrderedDict
import pytest
import wmul_test_utils

from wmul_rivendell.FilterCartReportForMusicScheduler import FilterCartReportForMusicScheduler
from wmul_rivendell.LoadCartDataDump import RivendellCart, CartType


cart_filter_params, cart_filter_ids = \
    wmul_test_utils.generate_true_false_matrix_from_list_of_strings(
        "cart_filter_params", ["fix_bad_header"]
    )

@pytest.fixture(scope="function", params=cart_filter_params, 
    ids=cart_filter_ids )
def setup_standard_cart_filter():
    mock_output_filename = "mock_output_filename"
    mock_rivendell_carts = "mock_rivendell_carts"

    cart_filter = FilterCartReportForMusicScheduler(
        rivendell_carts=mock_rivendell_carts,
        output_filename=mock_output_filename,
        desired_field_list=[],
        use_trailing_comma=False
    )

    return wmul_test_utils.make_namedtuple(
        "setup_standard_cart_filter",
        mock_output_filename=mock_output_filename,
        mock_rivendell_carts=mock_rivendell_carts,
        cart_filter=cart_filter
    )

@pytest.fixture(scope="function")
def defined_rivendell_carts():
    rivendell_cart_1_1 = RivendellCart(cart_number='1', cut_number='1', type=CartType.Audio, group_name='TEMPORIBUS', title='perferendis optio adipisci odit', artist='nesciunt repellendus', album='', year='1951', isrc='', isci='', label='', client='saepe architecto sequi alias', agency='', publisher='', composer='facere odio magnam odit', conductor='', song_id='', user_defined='', description='perferendis pariatur eligendi dolorem', outcue='', filename='000001_001.wav', length=':07', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes='')
    rivendell_cart_2_1 = RivendellCart(cart_number='2', cut_number='1', type=CartType.Audio, group_name='TEMPORIBUS', title='saepe recusandae deleniti laborum', artist='quisquam natus', album='', year='1965', isrc='', isci='', label='', client='enim ut qui veniam', agency='', publisher='', composer='omnis voluptatum accusantium saepe', conductor='', song_id='', user_defined='', description='praesentium eaque deleniti tempore', outcue='', filename='000002_001.wav', length=':07', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes='')
    rivendell_cart_2_2 = RivendellCart(cart_number='2', cut_number='2', type=CartType.Audio, group_name='TEMPORIBUS', title='aperiam sed totam reiciendis', artist='recusandae nesciunt', album='', year='1962', isrc='', isci='', label='', client='fugiat saepe quam libero', agency='', publisher='', composer='repellat non in fuga', conductor='', song_id='', user_defined='', description='deserunt repellat odit velit', outcue='', filename='000002_002.wav', length=':11', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes='')
    rivendell_cart_6_2 = RivendellCart(cart_number='6', cut_number='2', type=CartType.Audio, group_name='TEMPORIBUS', title='tempora libero earum repudiandae', artist='modi quo', album='', year='1953', isrc='', isci='', label='', client='iure nobis illum quia', agency='', publisher='', composer='fugiat beatae perspiciatis dicta', conductor='', song_id='', user_defined='', description='tempore voluptates tenetur dolorem', outcue='', filename='000006_002.wav', length=':11', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes='')
    rivendell_cart_101_1 = RivendellCart(cart_number='101', cut_number='1', type=CartType.Audio, group_name='TEMPORIBUS', title='aliquam omnis quisquam enim', artist='nobis sit', album='', year='2003', isrc='', isci='', label='', client='culpa quas minus reiciendis', agency='', publisher='', composer='ipsam cumque nulla harum', conductor='', song_id='', user_defined='', description='deserunt natus quidem at', outcue='', filename='000101_001.wav', length=':05', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes='')
    rivendell_cart_500_1 = RivendellCart(cart_number='500', cut_number='1', type=CartType.Audio, group_name='VOLUPTATIB', title='voluptate commodi neque aperiam', artist='reiciendis rem', album='', year='1969', isrc='', isci='', label='', client='assumenda quasi maxime cum', agency='', publisher='', composer='doloremque provident dolorum mollitia', conductor='', song_id='', user_defined='', description='ut praesentium sit quidem', outcue='', filename='000500_001.wav', length=':08', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes='')
    rivendell_cart_970000_1 = RivendellCart(cart_number='970000', cut_number='1', type=CartType.Macro, group_name='ALIQUAM', title='quasi necessitatibus quo vero', artist='sapiente fuga', album='', year='1994', isrc='', isci='', label='', client='distinctio culpa perspiciatis quam', agency='', publisher='', composer='explicabo magni temporibus provident', conductor='', song_id='', user_defined='', description='adipisci officia impedit dolores', outcue='', filename='', length=':00', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes='')
    rivendell_cart_970001_1 = RivendellCart(cart_number='970001', cut_number='1', type=CartType.Macro, group_name='ALIQUAM', title='porro eos dolores fuga', artist='rem dolor', album='', year='2015', isrc='', isci='', label='', client='omnis molestias nesciunt cumque', agency='', publisher='', composer='adipisci quidem consequuntur quia', conductor='', song_id='', user_defined='', description='provident perspiciatis quisquam ea', outcue='', filename='', length=':00', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes='')
    return wmul_test_utils.make_namedtuple(
        "defined_carts",
        rivendell_cart_1_1=rivendell_cart_1_1,
        rivendell_cart_2_1=rivendell_cart_2_1,
        rivendell_cart_2_2=rivendell_cart_2_2,
        rivendell_cart_6_2=rivendell_cart_6_2,
        rivendell_cart_101_1=rivendell_cart_101_1,
        rivendell_cart_500_1=rivendell_cart_500_1,
        rivendell_cart_970000_1=rivendell_cart_970000_1,
        rivendell_cart_970001_1=rivendell_cart_970001_1
    )


all_fields = [
        "cart_number",
        "cut_number",
        "type",
        "group_name",
        "title",
        "artist",
        "album",
        "year",
        "isrc",
        "isci",
        "label",
        "client",
        "agency",
        "publisher",
        "composer",
        "conductor",
        "song_id",
        "user_defined",
        "description",
        "outcue",
        "filename",
        "length",
        "start_point",
        "end_point",
        "segue_start_point",
        "segue_end_point",
        "hook_start_point",
        "hook_end_point",
        "talk_start_point",
        "talk_end_point",
        "fadeup_point",
        "fadedown_point",
        "sched_codes"
]


desired_field_list_parameters = [
    all_fields,
    [
        "cart_number",
        "cut_number",
        "type",
        "group_name",
        "title",
        "artist",
        "song_id",
        "user_defined",
        "description",
        "outcue",
        "filename",
        "length",
        "start_point",
        "end_point",
        "talk_start_point",
        "talk_end_point",
        "fadeup_point",
        "fadedown_point",
        "sched_codes"
    ],
    [
        "cart_number",
        "cut_number",
        "type"
    ]
]


@pytest.mark.parametrize("desired_field_list", desired_field_list_parameters)
def test__remove_unwanted_fields_fields_removed(
        setup_standard_cart_filter, 
        defined_rivendell_carts, 
        desired_field_list
    ):

    cart_filter = setup_standard_cart_filter.cart_filter

    rivendell_carts_for_test = [
        defined_rivendell_carts.rivendell_cart_1_1,
        defined_rivendell_carts.rivendell_cart_2_1,
        defined_rivendell_carts.rivendell_cart_2_2,
        defined_rivendell_carts.rivendell_cart_6_2,
        defined_rivendell_carts.rivendell_cart_101_1,
        defined_rivendell_carts.rivendell_cart_500_1
    ]

    cart_filter.desired_field_list = desired_field_list

    result_carts = cart_filter._remove_unwanted_fields(
        rivendell_carts=rivendell_carts_for_test
    )

    assert len(result_carts) == len(rivendell_carts_for_test)

    for rc in result_carts:
        assert desired_field_list == list(rc.keys())


def test__remove_unwanted_fields_fields_copied_correctly(
        setup_standard_cart_filter, 
        defined_rivendell_carts
    ):

    cart_filter = setup_standard_cart_filter.cart_filter

    rivendell_carts_for_test = [
        defined_rivendell_carts.rivendell_cart_1_1
    ]
    cart_filter.desired_field_list = all_fields

    result_carts = cart_filter._remove_unwanted_fields(
        rivendell_carts=rivendell_carts_for_test
    )

    assert len(result_carts) == len(rivendell_carts_for_test)

    cart_under_test = rivendell_carts_for_test[0]

    result_cart = result_carts[0]
    result_cart_keys = result_cart.keys()

    invalid_field_name = "INVALID FIELD NAME"

    for field in all_fields:
        assert field in result_cart_keys
        original_value_of_field = getattr(cart_under_test, field, 
                                          invalid_field_name)
        assert original_value_of_field != invalid_field_name
        assert result_cart[field] == original_value_of_field


@pytest.fixture()
def defined_music_scheduler_carts():
    cart_number_field_name = "cart_number"
    cut_number_field_name = "cut_number"
    group_name_field_name = "group_name"
    title_field_name = "title"

    rc_1_1_cart_number = '1'
    rc_1_1_cut_number = '1'
    rc_1_1_group_name = "TEMPORIBUS"
    rc_1_1_title = "perferendis optio adipisci odit"

    rc_1_1 = OrderedDict()
    rc_1_1[cart_number_field_name] = rc_1_1_cart_number
    rc_1_1[cut_number_field_name] = rc_1_1_cut_number
    rc_1_1[group_name_field_name] = rc_1_1_group_name
    rc_1_1[title_field_name] = rc_1_1_title

    rc_1_1_line = rc_1_1_cart_number + "," + rc_1_1_cut_number + "," + rc_1_1_group_name + "," + rc_1_1_title

    rc_6_2_cart_number = '6'
    rc_6_2_cut_number = '2'
    rc_6_2_group_name = "TEMPORIBUS"
    rc_6_2_title = "tempora libero earum repudiandae"

    rc_6_2 = OrderedDict()
    rc_6_2[cart_number_field_name] = rc_6_2_cart_number
    rc_6_2[cut_number_field_name] = rc_6_2_cut_number
    rc_6_2[group_name_field_name] = rc_6_2_group_name
    rc_6_2[title_field_name] = rc_6_2_title

    rc_6_2_line = rc_6_2_cart_number + "," + rc_6_2_cut_number + "," + rc_6_2_group_name + "," + rc_6_2_title

    rc_101_1_cart_number = '101'
    rc_101_1_cut_number = '1'
    rc_101_1_group_name = "LEGATEMPORIBUSL_ID"
    rc_101_1_title = "aliquam omnis quisquam enim"

    rc_101_1 = OrderedDict()
    rc_101_1[cart_number_field_name] = rc_101_1_cart_number
    rc_101_1[cut_number_field_name] = rc_101_1_cut_number
    rc_101_1[group_name_field_name] = rc_101_1_group_name
    rc_101_1[title_field_name] = rc_101_1_title

    rc_101_1_line = rc_101_1_cart_number + "," + rc_101_1_cut_number + "," + rc_101_1_group_name + "," + rc_101_1_title

    rc_500_1_cart_number = '500'
    rc_500_1_cut_number = '1'
    rc_500_1_group_name = "VOLUPTATIB"
    rc_500_1_title = "voluptate commodi neque aperiam"

    rc_500_1 = OrderedDict()
    rc_500_1[cart_number_field_name] = rc_500_1_cart_number
    rc_500_1[cut_number_field_name] = rc_500_1_cut_number
    rc_500_1[group_name_field_name] = rc_500_1_group_name
    rc_500_1[title_field_name] = rc_500_1_title

    rc_500_1_line = rc_500_1_cart_number + "," + rc_500_1_cut_number + "," + rc_500_1_group_name + "," + rc_500_1_title

    return wmul_test_utils.make_namedtuple(
        "defined_music_scheduler_carts",
        rc_1_1=rc_1_1,
        rc_1_1_line=rc_1_1_line,
        rc_6_2=rc_6_2,
        rc_6_2_line=rc_6_2_line,
        rc_101_1=rc_101_1,
        rc_101_1_line=rc_101_1_line,
        rc_500_1=rc_500_1,
        rc_500_1_line=rc_500_1_line
    )


def test__export_carts_to_csv_trailing_comma(fs, setup_standard_cart_filter, 
                                             defined_music_scheduler_carts):
    cart_filter = setup_standard_cart_filter.cart_filter

    import pathlib
    output_file_path = pathlib.Path(r"\fake_output_file.csv")
    assert not output_file_path.exists()

    cart_filter.output_filename = output_file_path
    cart_filter.use_trailing_comma = True

    music_scheduler_carts = [
        defined_music_scheduler_carts.rc_1_1,
        defined_music_scheduler_carts.rc_6_2,
        defined_music_scheduler_carts.rc_101_1,
        defined_music_scheduler_carts.rc_500_1
    ]

    cart_filter._export_carts_to_csv(music_scheduler_carts)

    assert output_file_path.exists()

    expected_file_contents = \
        defined_music_scheduler_carts.rc_1_1_line + ",\r\n" + \
        defined_music_scheduler_carts.rc_6_2_line + ",\r\n" + \
        defined_music_scheduler_carts.rc_101_1_line + ",\r\n" + \
        defined_music_scheduler_carts.rc_500_1_line + ",\r\n"

    result_file_contents = open(output_file_path, newline="", mode="rt", 
                                errors="replace").read()

    assert expected_file_contents == result_file_contents


def test__export_carts_to_csv_no_trailing_comma(fs, setup_standard_cart_filter, 
                                                defined_music_scheduler_carts):
    cart_filter = setup_standard_cart_filter.cart_filter

    import pathlib
    output_file_path = pathlib.Path(r"\fake_output_file.csv")
    assert not output_file_path.exists()

    cart_filter.output_filename = output_file_path
    cart_filter.use_trailing_comma = False

    music_scheduler_carts = [
        defined_music_scheduler_carts.rc_1_1,
        defined_music_scheduler_carts.rc_6_2,
        defined_music_scheduler_carts.rc_101_1,
        defined_music_scheduler_carts.rc_500_1
    ]

    cart_filter._export_carts_to_csv(music_scheduler_carts)

    assert output_file_path.exists()

    expected_file_contents = \
        defined_music_scheduler_carts.rc_1_1_line + "\r\n" + \
        defined_music_scheduler_carts.rc_6_2_line + "\r\n" + \
        defined_music_scheduler_carts.rc_101_1_line + "\r\n" + \
        defined_music_scheduler_carts.rc_500_1_line + "\r\n"

    result_file_contents = open(output_file_path, newline="", mode="rt", 
                                errors="replace").read()

    assert expected_file_contents == result_file_contents


@pytest.fixture(scope="function")
def setup_run_script(caplog, mocker):
    mock_output_filename = "mock_output_filename"

    mock_rivendell_carts = "mock_rivendell_carts"

    mock_music_scheduler_carts = "mock_music_scheduler_cuts"
    mock_remove_unwanted_fields = mocker.Mock(return_value=mock_music_scheduler_carts)
    mocker.patch(
        "wmul_rivendell.FilterCartReportForMusicScheduler.FilterCartReportForMusicScheduler._remove_unwanted_fields",
        mock_remove_unwanted_fields
    )

    mock_export_carts_to_csv = mocker.Mock()
    mocker.patch(
        "wmul_rivendell.FilterCartReportForMusicScheduler.FilterCartReportForMusicScheduler._export_carts_to_csv",
        mock_export_carts_to_csv
    )

    cart_filter = FilterCartReportForMusicScheduler(
        rivendell_carts=mock_rivendell_carts,
        output_filename=mock_output_filename,
        desired_field_list=[],
        use_trailing_comma=True
    )

    cart_filter.run_script()

    return wmul_test_utils.make_namedtuple(
        "setup_run_script",
        mock_rivendell_carts=mock_rivendell_carts,
        mock_remove_unwanted_fields=mock_remove_unwanted_fields,
        mock_music_scheduler_carts=mock_music_scheduler_carts,
        mock_export_carts_to_csv=mock_export_carts_to_csv,
        caplog_text=caplog.text
    )


def test_run_script_remove_unwanted_fields_called_correctly(setup_run_script):
    setup_run_script.mock_remove_unwanted_fields.assert_called_once_with(
        setup_run_script.mock_rivendell_carts
    )


def test_run_script_export_carts_to_csv_called_correctly(setup_run_script):
    setup_run_script.mock_export_carts_to_csv.assert_called_once_with(setup_run_script.mock_music_scheduler_carts)



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['wmul_rivendell']

package_data = \
{'': ['*']}

package_dir = \
{'': 'src'}

install_requires = \
['wmul_logger>=0.5.1',
 'wmul_emailer>=0.5.0',
 'click',
 'cachetools',
 'numpy',
 'pandas',
 'openpyxl']

extras_require = \
{":python_version < '3.7.0'": ['mysql-connector-python<8.0.29',
                               'protobuf<=3.19.4',
                               'dataclasses'],
 ":python_version > '3.7.0'": ['mysql-connector-python>8.0',
                               'protobuf>=3.19.0'],
 'test': ['pytest', 'pytest-mock', 'pyfakefs', 'wmul_test_utils>=0.3.1'],
 'tox': ['tox']}

entry_points = \
{'console_scripts': ['wmul_rivendell = wmul_rivendell.cli:wmul_rivendell_cli']}

setup(name='wmul_rivendell',
      version='0.13.1',
      description='Various scripts to help use Rivendell Radio Automation 3.4.1+',
      author=None,
      author_email='Michael Stanley <stanley50@marshall.edu>',
      url=None,
      packages=packages,
      package_data=package_data,
      package_dir=package_dir,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
     )



"""
@Author = 'Michael Stanley'

============ Change Log ============
2025-Jan-03 = Created

============ License ============
Copyright (C) 2025 Michael Stanley

This file is part of wmul_rivendell.

wmul_rivendell is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License as published by the Free 
Software Foundation, either version 3 of the License, or (at your option) any 
later version.

wmul_rivendell is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
wmul_rivendell. If not, see <https://www.gnu.org/licenses/>. 
"""
from click.testing import CliRunner
import pytest

from wmul_rivendell import cli
from wmul_test_utils import generate_true_false_matrix_from_list_of_strings

database_statistic_params, database_statistic_ids = \
    generate_true_false_matrix_from_list_of_strings(
        "database_statistic",
        [
            "include_all_cuts",
            "exclude_groups",
            "provide_smallest_stdev",
            "provide_minimum_population",
            "provide_lower_bound_multiple",
            "provide_upper_bound_multiple",
            "write_limits",
            "write_full_statistics"
        ]

    )

@pytest.mark.parametrize("params", database_statistic_params, ids=database_statistic_ids)
def test_database_statistics(fs, params, mocker, caplog):
    from pathlib import Path
    mock_rivendell_cart_filename = "/test/mock_rivendell_cart_filename.txt"
    fs.create_file(mock_rivendell_cart_filename)
    mock_output_filename = "/test/mock_output_filename"
    expected_output_filename = Path(mock_output_filename)

    mock_rivendell_carts = "mock_rivendell_carts"
    mock_load_carts = mocker.Mock(return_value=mock_rivendell_carts)

    mock_load_cart_data_dump_object = mocker.Mock(load_carts=mock_load_carts)
    mock_load_cart_data_dump_constructor = mocker.patch(
        "wmul_rivendell.cli.LoadCartDataDump",
        return_value=mock_load_cart_data_dump_object,
        autospec=True
    )

    mock_stats_limits_object = mocker.Mock()
    mock_stats_limits_constructor = mocker.patch(
        "wmul_rivendell.cli.StatisticsLimits",
        return_value=mock_stats_limits_object,
        autospec=True
    )

    mock_database_statistics_object = mocker.Mock()

    mock_database_statistics_constructor = mocker.patch(
        "wmul_rivendell.cli.DatabaseStatistics",
        return_value=mock_database_statistics_object,
        autospec=True
    )

    cli_args = [
            mock_rivendell_cart_filename,
            mock_output_filename,
        ]

    if params.include_all_cuts:
        cli_args.append("--include_all_cuts")

    if params.exclude_groups:
        exclude_groups_file_contents = "LAUDANTIUM\nASPERIORES\n"
        mock_exclude_groups_filename = "/test/mock_exclude_groups_filename.txt"
        fs.create_file(mock_exclude_groups_filename, contents=exclude_groups_file_contents)
        cli_args.append("--excluded_groups_file_name")
        cli_args.append(mock_exclude_groups_filename)
        expected_exclude_groups = ["LAUDANTIUM", "ASPERIORES"]
    else:
        expected_exclude_groups = []

    if params.provide_smallest_stdev:
        expected_smallest_stdev = 31
        cli_args.append("--smallest_stdev")
        cli_args.append(expected_smallest_stdev)
    else:
        expected_smallest_stdev = 15

    if params.provide_minimum_population:
        expected_minimum_population = 10
        cli_args.append("--minimum_population")
        cli_args.append(expected_minimum_population)
    else:
        expected_minimum_population = 4

    if params.provide_lower_bound_multiple:
        expected_lower_bound_multiple = 2
        cli_args.append("--lower_bound_multiple")
        cli_args.append(expected_lower_bound_multiple)
    else:
        expected_lower_bound_multiple = 1.5

    if params.provide_upper_bound_multiple:
        expected_upper_bound_multiple = 4.5
        cli_args.append("--upper_bound_multiple")
        cli_args.append(expected_upper_bound_multiple)
    else:
        expected_upper_bound_multiple = 3.0

    expected_write_limits = params.write_limits
    if params.write_limits:
        cli_args.append("--write_limits")

    expected_write_full_statistics = params.write_full_statistics
    if params.write_full_statistics:
        cli_args.append("--write_full_statistics")


    runner = CliRunner()
    result = runner.invoke(
        cli.database_statistics,
        cli_args
    )

    assert result.exit_code == 0

    mock_stats_limits_constructor.assert_called_once_with(
        smallest_stdev=expected_smallest_stdev,
        minimum_population_for_outliers=expected_minimum_population,
        lower_bound_multiple=expected_lower_bound_multiple,
        upper_bound_multiple=expected_upper_bound_multiple
    )

    mock_load_cart_data_dump_constructor.assert_called_once_with(
        rivendell_cart_data_filename=mock_rivendell_cart_filename,
        include_macros=False,
        include_all_cuts=params.include_all_cuts,
        excluded_group_list=expected_exclude_groups,
    )

    mock_load_carts.assert_called_once_with()

    mock_database_statistics_constructor.assert_called_once_with(
        rivendell_carts=mock_rivendell_carts,
        output_filename=expected_output_filename,
        stats_limits=mock_stats_limits_object,
        write_limits=expected_write_limits,
        write_full_statistics=expected_write_full_statistics
    )

    mock_database_statistics_object.run_script.assert_called_once_with()



"""
@Author = 'Michael Stanley'

This script takes the "Cart Data Dump (CSV)" from RD Library and eliminates data fields that the music scheduler
cannot use. It does this because some music schedulers, such as Natural Music 5, cannot import the full data dump.

============ Change Log ============
2025-Jan-03 = Change the way the trailing comma is added. Python 3.13 doesn't allow a comma to be included in the line 
              terminator. Instead, a dummy additional field is added to the end of the record.

2023-May-25 = Refactor to remove RivendellCart, CartType, and the logic for 
                loading and filtering the carts into LoadCartDataDump. This 
                refactoring will allow DatabaseStatistics to re-use the code.

2023-Jan-20 = Back-insert items into change log.
              Change license from GPLv2 to GPLv3.

2023-Jan-19 = Modify to accomodate a bug in Rivendell 3.6.4-3.6.6 where two 
                fields in the header are malformed.

2022-May-06 = Update documentation.

2021-May-26 = Change RivendellCart to keep cart_number, cut_number, and year as
                strings. 
              Refactor from_csv to from_dict, to more accurately reflect the 
                way it works.
              Refactor to make the removal of macro carts into a method, for
                easier unit testing.

2020-Jun-25 = Created.

============ License ============
Copyright (C) 2020-2023, 2025 Michael Stanley

This file is part of wmul_rivendell.

wmul_rivendell is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License as published by the Free 
Software Foundation, either version 3 of the License, or (at your option) any 
later version.

wmul_rivendell is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
wmul_rivendell. If not, see <https://www.gnu.org/licenses/>. 
"""
import csv
from collections import OrderedDict
from dataclasses import dataclass
from pathlib import Path


import wmul_logger

_logger = wmul_logger.get_logger()


@dataclass
class FilterCartReportForMusicScheduler:
    rivendell_carts: list
    output_filename: Path
    desired_field_list: list
    use_trailing_comma: bool

    def _remove_unwanted_fields(self, rivendell_carts):
        music_scheduler_carts = []
        for cart in rivendell_carts:
            music_scheduler_cart = OrderedDict()
            for desired_field in self.desired_field_list:
                music_scheduler_cart[desired_field] = getattr(cart, desired_field.lower(), "INVALID FIELD NAME")
            music_scheduler_carts.append(music_scheduler_cart)
        return music_scheduler_carts

    def _export_carts_to_csv(self, music_scheduler_carts):
        fieldnames = list(music_scheduler_carts[0].keys())
        if self.use_trailing_comma:
            fieldnames.append("Placeholder")
        with open(str(self.output_filename), newline="", mode="wt", errors="replace") as music_scheduler_file:
            natural_music_writer = csv.DictWriter(music_scheduler_file, fieldnames,
                                                  dialect="excel")
            natural_music_writer.writerows(music_scheduler_carts)

    def run_script(self):
        _logger.debug(f"Starting run_script with {self}")
        
        music_scheduler_carts = self._remove_unwanted_fields(self.rivendell_carts)
        self._export_carts_to_csv(music_scheduler_carts)
